# -*- coding: utf-8 -*-
"""
Created on Sat Mar  4 10:57:14 2017

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import os
import numpy as np
from re import findall as refindall


from grapa.graph import Graph
from grapa.curve import Curve



class GraphMCAfit(Graph):
    
    FILEIO_GRAPHTYPE = 'XRF fit areas'
    
    AXISLABELS = [['Data', '', None], ['Value', '', None]]
    
    @classmethod
    def isFileReadable(cls, fileName, fileExt, line1='', **kwargs):
        if fileExt == '.html' and line1[0:24] == '<HTML><HEAD><TITLE>PyMCA':
            return True
        return False

    def readDataFromFile(self, attributes, **kwargs):
        filenam_, fileext = os.path.splitext(self.filename)
        sample = filenam_.split('/')[-1].split('.mca')[0]
        f = open(self.filename, 'r')
        content = f.read()
        f.close()
        tmp = np.array([])
        content = content.replace(' align="left"', '').replace(' align="right"', '').replace(' bgcolor=#E6F0F9', '').replace(' bgcolor=#E6F0F9', '').replace(' bgcolor=#FFFFFF', '').replace(' bgcolor=#FFFACD', '').replace('  ', ' ').replace('<td >', '<td>').replace('<tr', '\n<tr').replace('<TR', '\n<TR').replace('<table', '\n<table').replace('\n\n', '\n')
        split = refindall('<tr><td>Cu</td><td>K</td><td>([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)</td><td>', content)
        tmp = np.append(tmp, float(split[0][0]))
        split = refindall('<tr><td>In</td><td>K</td><td>([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)</td><td>', content)
        tmp = np.append(tmp, float(split[0][0]))
        split = refindall('<tr><td>Ga</td><td>Ka</td><td>([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)</td><td>', content)
        tmp = np.append(tmp, float(split[0][0]))
        split = refindall('<tr><td>Se</td><td>K</td><td>([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)</td><td>', content)
        tmp = np.append(tmp, float(split[0][0]))
        tmp = np.append(tmp, 180) # time - assumed here, no value reading
    ## XRF calibration: MIGHT WANT TO UPDATE THAT!!!
        val_Cu = tmp[0] / 43434 * 22.56
        val_In = tmp[1] / 10992 * 13.66
        val_Ga = tmp[2] / 29578 * 11.41
        val_Se = tmp[3] /195460 * 52.40
    ## END OF XRF calibration: MIGHT WANT TO UPDATE THAT!!!
        val_sum = val_Cu + val_In + val_Ga + val_Se
        val_Cu /= val_sum
        val_In /= val_sum
        val_Ga /= val_sum
        val_Se /= val_sum
        tmp = np.append(tmp, val_Ga / (val_Ga + val_In)) # x calculation
        tmp = np.append(tmp, val_Cu / (val_Ga + val_In)) # y calculation
        tmp = np.append(tmp, tmp[0] * 450 / 10000000) # D calculation
        self.data.append(Curve(np.append(np.arange(tmp.size), tmp).reshape((2, tmp.size)), {}))
        c = self.curve(-1)
        c.update(attributes)
        c.update({'XRF fit Cu [fitarea/s]': float(tmp[0]) / float(tmp[4])})
        c.update({'XRF fit In [fitarea/s]': float(tmp[1]) / float(tmp[4])})
        c.update({'XRF fit Ga [fitarea/s]': float(tmp[2]) / float(tmp[4])})
        c.update({'XRF fit Se [fitarea/s]': float(tmp[3]) / float(tmp[4])})
        c.update({'XRF fit time [s]':float(tmp[4])})
        c.update({'XRF fit x [ ]'  : float(tmp[5])})
        c.update({'XRF fit y [ ]'  : float(tmp[6])})
        c.update({'XRF fit D [um]' : float(tmp[7])})
        # other info
        self.headers.update({'sample': sample, 'collabels': ['Element [ ]', 'Fit area / value [a.u.]']})
        self.update({'xlabel': self.formatAxisLabel(GraphMCAfit.AXISLABELS[0]),
                     'ylabel': self.formatAxisLabel(GraphMCAfit.AXISLABELS[1]),
                     'xtickslabels': [[0, 1, 2, 3, 4, 5, 6, 7],
                                      ['Cu', 'In', 'Ga', 'Se', 'Acq. time', 'GGI', 'CGI', 'D']]})
        c.update({'linespec': 'd'})
        c.update({'label': c.getAttribute('label').split('.')[0]})
        print('   XRF composition: GGI', '{:1.3f}'.format(tmp[5]),
              ', CGI', '{:1.3f}'.format(tmp[6]),
              ', D', '{:1.3f}'.format(tmp[7]),'(calibration x=0.3).')
        GaSe = 1.63   * tmp[2] / tmp[3]
        CuSe = 2.0661 * tmp[0] / tmp[3]
        GGI_ = 2 * GaSe / (1 + 1/3 * (1 - 2 * CuSe))
        CGI_ = 2 * CuSe / (1 + 1/3 * (1 - 2 * CuSe))
        D_ = 9.50E-06 * tmp[3]
        print('   XRF Se ratios composition: GGI', '{:1.3f}'.format(GGI_),
              ', CGI', '{:1.3f}'.format(CGI_), 'D', '{:1.3f}'.format(D_),
              '(calibration Ga/Se, Cu/Se)')



# -*- coding: utf-8 -*-
"""
Created on Sat Mar  4 10:44:14 2017

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import os
import numpy as np


from grapa.graph import Graph
from grapa.curve import Curve

class GraphMBElog(Graph):
    
    FILEIO_GRAPHTYPE = 'Small MBE log file'
    
    AXISLABELS = [['Time', 't', 'min'], ['Substrate heating', '', 'a.u.']]
    
    @classmethod
    def isFileReadable(cls, fileName, fileExt, line1='', **kwargs):
        if fileExt == '.txt' and line1 == 'Time	Value':
            return True
        return False

    def readDataFromFile(self, attributes, **kwargs):
        """ Reads MBE log file. Curve is standard xyCurve. """
        fileName, fileext = os.path.splitext(self.filename)
        sample = (fileName.split('/')[-1]).split('_')[0]
        data = np.genfromtxt(self.filename, skip_header=1, delimiter='\t', 
                             usecols=[0, 1], invalid_raise=False)
        self.data.append(Curve(np.transpose(data), attributes))
        self.headers.update({'collabels': ['Time [min]',
                                           'Substrate heating [a.u.]']})
        self.update({'xlabel': self.formatAxisLabel(GraphMBElog.AXISLABELS[0]),
                     'ylabel': self.formatAxisLabel(GraphMBElog.AXISLABELS[1])})
        self.curve(-1).update({'label': sample})





# -*- coding: utf-8 -*-
"""
@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import numpy as np
import os
import sys

import glob
from copy import deepcopy
from re import search as research
import matplotlib.pyplot as plt

path = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', '..'))
if path not in sys.path:
    sys.path.append(path)

from grapa.graph import Graph
from grapa.graphIO import GraphIO
from grapa.database import Database
from grapa.mathModule import is_number, roundSignificant, stringToVariable
from grapa.colorscale import Colorscale
from grapa.curve import Curve
from grapa.curve_subplot import Curve_Subplot

from grapa.datatypes.curveJV import GraphJV
from grapa.datatypes.graphJVDarkIllum import GraphJVDarkIllum


# prompt for folder
def promptForFolder ():
    # Start->Execute, cmd, type "pip install tkinter"
    from tkinter.filedialog import askdirectory
    path = askdirectory()
    return path



# auxiliary function
def dictToListSorted (d) :
    l = []
    for key in d :
        l.append(key)
    l.sort()
    return l

# auxiliary class - add some useful tools to the database class
class areaDB (Database) :
    def __init__ (self, folder, sample) :
        self.colIdx = 0
        self.folder = folder
        self.sample = sample
        # different possible names for area databases
        test = []
        test.append([os.path.join(folder, sample+'.txt'), ''])
        test.append([os.path.join(folder, sample+'_area.txt'), ''])
        test.append([os.path.join(folder, sample+'_areas.txt'), ''])
        test.append([os.path.join(folder, 'area.txt'), ''])
        test.append([os.path.join(folder, 'area.xlsx'), sample])
        test.append([os.path.join(folder, 'areas.txt'), ''])
        test.append([os.path.join(folder, 'areas.xlsx'), sample])
        self.flag = False
        self.flagCheat = False
        for t in test :
            print ('Area database: look in file ',t[0])
            graph = Graph(t[0], complement=t[1], silent=True, config=None)
            if graph.length() == 0 :
                continue
            else :
                try :
                    # try convert it in database
                    Database.__init__ (self, graph)
                    # identify suitable column
                    for col in self.colLabels :
                        if col.find ('area') > -1 :
                            self.colIdx = col
                    self.flag = True
                    print ('   database file parsed and area identified.')
                except Exception:
                    print ('   database file data could not be interpreted correctly.')
#                    print (e)
                    continue
                break # break if success
        if not self.flag :
            print ('areaDB: cannot find area database file.')
    
    def getArea (self, cell):
        if not self.flag : # if database could not be opened
            if self.flagCheat:
                if cell in self.data:
                    return self.data[cell]
            return np.nan
        out = self.value (self.colIdx, cell)
        if np.isnan (out) :
            out = self.value (self.colIdx, self.sample+' '+cell)
        if np.isnan (out) :
            print ('areaDB getArea: row "', cell, '"not found (', self.rowLabels,')')
#        print ('area cell',cell,':', out, '(',self.sample,')')
        return out
    
    def setArea (self, cell, value):
        if self.flag:
            self.setValue(self.colIdx, cell, value, silent=True)
        else:
            self.flagCheat = True
            if not hasattr(self, 'data'):
                self.data = {}
            self.data.update({cell: value})
        


# main function
def processJVfolder (folder, ylim=[-50,150], sampleName='', fitDiodeWeight=0, groupCell=True, figAx=None, pltClose=True, newGraphKwargs={}):
    print('Script processJV folder initiated. Data processing can last a few seconds.')
    newGraphKwargs = deepcopy(newGraphKwargs)
    newGraphKwargs.update({'silent': True})
    if figAx is not None:
        pltClose = False
    path = os.path.join(folder, '*.txt')
    
    cellDict = {}
    sampleAreaDict = {}
    outFlag = True
    out0 = ''
    
    graph = Graph('', **newGraphKwargs)
    for file in glob.glob(path):
        fileName, fileExt = os.path.splitext(file)
        fileExt = fileExt.lower()
        line1, line2, line3 = GraphIO.readDataFileLine123(filename=file)
        if GraphJV.isFileReadable(fileName, fileExt, line1=line1, line2=line2, line3=line3):
            try:
                graphTmp = Graph(file, **newGraphKwargs)
            except IndexError as e:
                continue # next file
        else:
            continue
        try:
            sample = graphTmp.curve(-1).sample()
            cell   = graphTmp.curve(-1).cell()
            dark   = graphTmp.curve(-1).darkOrIllum (ifText=True)
            measId = graphTmp.curve(-1).measId()
        except Exception:
            continue # go to next file
        
        print ('file', file)
        if sample == '' or cell == '':
            print('WARNING: cannot identify sample (', sample, ') or cell (', cell, ').')

        if sample != '' and cell != '' :
            if sample not in cellDict:
                cellDict.update ({sample: {}})
            if cell not in cellDict[sample]:
                cellDict[sample].update ({cell: {}})
            if dark not in cellDict[sample][cell]:
                cellDict[sample][cell].update ({dark: {}})
            if sample not in sampleAreaDict:
                sampleAreaDict.update ({sample: areaDB(folder, sample)})
            # open JV file again to correct for cell area
            area = sampleAreaDict[sample].getArea(cell)
            if np.isnan(area) and is_number(graphTmp.curve(0).getAttribute('Acquis soft Cell area')) and not np.isnan(graphTmp.curve(0).getAttribute('Acquis soft Cell area')):
                sampleAreaDict[sample].setArea(cell, graphTmp.curve(0).getAttribute('Acquis soft Cell area'))
                area = sampleAreaDict[sample].getArea(cell)
            cellDict[sample][cell][dark].update({measId: file})
            print (sample, 'cell', cell, 'area', area, 'cm2 (acquired as', graphTmp.curve(0).getAttribute('Acquis soft Cell area'), ')')

        if outFlag :
            out0 = out0 + graphTmp.curve(-1).printShort(header=True)
            outFlag = False


    # sweep through files, identify pairs
    listSample = dictToListSorted(cellDict)
    for s in listSample :
        out      = 'Sample\t' + s + '\n' + 'label\t' + s.replace('_','\\n') + '\n' + deepcopy(out0)
        outIllum = 'Sample\t' + s + '\n' + 'label\t' + s.replace('_','\\n') + '\n' + deepcopy(out0)
        outDark  = 'Sample\t' + s + '\n' + 'label\t' + s.replace('_','\\n') + '\n' + deepcopy(out0)
        listCell = dictToListSorted(cellDict[s])
        for c in listCell :
            listDarkIllum = dictToListSorted(cellDict[s][c])
            # if want to process each emasureemnt independently
            if not groupCell:
                for d in listDarkIllum:
                    listGraph = dictToListSorted(cellDict[s][c][d])
                    for m in listGraph:
                        filesave = 'export_' + s + '_' + c + '_' + m + ('_'+d).replace('_illum','')
                        print ('Graph saved as', filesave)
                        graph = GraphJVDarkIllum (cellDict[s][c][d][m], '', area=sampleAreaDict[s].getArea(c), complement={'ylim':ylim, 'saveSilent': True, '_fitDiodeWeight': fitDiodeWeight}, **newGraphKwargs)
                        out = out + graph.printShort()
                        filesave = os.path.join(folder, filesave)
                        graph.plot(filesave=filesave, figAx=figAx, pltClose=pltClose)
            else:
                # if want to restrict 1 msmt dark + 1 illum per cell
                if len(listDarkIllum) > 2 :
                    print ('test WARNING sorting JV files.')
                # only dark on only illum measurement
                if len(listDarkIllum) == 1 :
                    d = listDarkIllum[0]
                    listGraph = dictToListSorted (cellDict[s][c][d])
                    m = listGraph[0]
                    filesave = 'export_' + s + '_' + c + '_' + m + ('_'+d).replace('_illum','')
                    print ('Graph saved as', filesave)
                    fileDark = cellDict[s][c][d][m] if listDarkIllum[0] == 'dark'  else ''
                    fileIllum= cellDict[s][c][d][m] if listDarkIllum[0] == 'illum' else ''
                    # create Graph file
                    graph = GraphJVDarkIllum(fileDark, fileIllum, area=sampleAreaDict[s].getArea(c), complement={'ylim':ylim, 'saveSilent': True, '_fitDiodeWeight': fitDiodeWeight}, **newGraphKwargs)
                    filesave = os.path.join(folder, filesave)
                    graph.plot(filesave=filesave, figAx=figAx, pltClose=pltClose)
                    # prepare output summary files
                    out = out + graph.printShort()
                    if listDarkIllum[0] == 'dark' :
                        outDark  = outDark  + graph.printShort()
                    else :
                        outIllum = outIllum + graph.printShort()
                    #if len(listGraph) > 1 :
                    #    msg = '.'.join([cellDict[s][c][d][m2] for m2 in cellDict[s][c][d]][1:])
                    #    print ('test WARNING: other files ignored (,',msg,')')
    
                # can identify pair of dark-illum files
                if len(listDarkIllum) == 2 :
                    listGraph = dictToListSorted (cellDict[s][c][listDarkIllum[0]])
                    filesave = 'export_' + s + '_' + c + '_' + listGraph[0]
                    fileDark = cellDict[s][c][listDarkIllum[0]][listGraph[0]]

                    listGraph = dictToListSorted (cellDict[s][c][listDarkIllum[1]])
                    filesave = filesave + '-' + listGraph[0]
                    fileIllum= cellDict[s][c][listDarkIllum[1]][listGraph[0]]

                    print ('Graph saved as', filesave)
                    # create Graph file
                    graph = GraphJVDarkIllum (fileDark, fileIllum, area=sampleAreaDict[s].getArea(c), complement={'ylim':ylim, 'saveSilent': True, '_fitDiodeWeight': fitDiodeWeight}, **newGraphKwargs)
                    filesave = os.path.join(folder, filesave)
                    graph.plot(filesave=filesave, figAx=figAx, pltClose=pltClose)
                    # prepare output summary files
                    out      = out      + graph.printShort()
                    outIllum = outIllum + graph.printShort(onlyIllum=True)
                    outDark  = outDark  + graph.printShort(onlyDark =True)
        # print sample summary
        filesave = 'export_' + s + '_summary' + '.txt'
        filesave = os.path.join(folder, filesave)
        #print('End of JV curves processing, showing summary file...')
        #print(out)
        print('Summary saved in file', filesave, '.')
        f = open(filesave, 'w')
        f.write(out)
        f.close()
        if groupCell:
            filesave = 'export_' + s + '_summary_dark' + '.txt'
            filesave = os.path.join(folder, filesave)
            f = open(filesave, 'w')
            f.write(outDark)
            f.close()
            processSampleCellsMap(filesave, figAx=figAx, pltClose=pltClose)
            filesave = 'export_' + s + '_summary_illum' + '.txt'
            filesave = os.path.join(folder, filesave)
            f = open(filesave, 'w')
            f.write(outIllum)
            f.close()
            processSampleCellsMap(filesave, figAx=figAx, pltClose=pltClose)
            print(writeFileAvgMax(filesave))
    print('Script processJV folder done.')
    return graph



def writeFileAvgMax(fileOrContent, filesave=None, withHeader=True, colSample=True):
    colOfInterest = ['Voc', 'Jsc', 'FF', 'Eff']
    if isinstance(fileOrContent, Graph):
        content = fileOrContent
        filename = content.getAttribute('sample').replace('\n','')
        if filename == '':
            filename = content.getAttribute('label')
    else:
        content = Graph(fileOrContent)
        filename = fileOrContent
#    print(content)
    colLbl = content.getAttribute('collabels')
    cols = []
    idxs = []
    for c in colOfInterest:
        cols.append([np.nan])
        idxs.append(np.nan)
        for i in range(len(colLbl)):
            if c in colLbl[i]:
                cols[-1] = content.curve(i).y()
                idxs[-1] = i
    out = ''
    if withHeader:
        if not colSample:
            out += 'filename\t' + filename + '\n' 
            out += 'Sample\t' + content.getAttribute('sample') + '\n'
        if colSample:
            out += '\t'
        # column headers
        out += 'Parameter average' + '\t'*len(colOfInterest)
        out += 'Best cell (eff.)' + '\t'*len(colOfInterest)
        out += 'Parameter median' + '\t'*len(colOfInterest)
        out += '\n'
        # column name
        if colSample:
            out += 'Sample\t'
        for i in range(3):
            for c in idxs:
                out += (colLbl[c] if c is not np.isnan(c) else '') + '\t'
        out += '\n'
    # averages
    if colSample:
        out += (content.headers['sample'] if 'sample' in content.headers else 'SOMETHING') + '\t'
    for c in cols:
        out += str(np.average(c)) + '\t'
    # best cell
    eff = None
    for i in range(len(colLbl)):
        if 'Eff' in colLbl[i]:
            eff = content.curve(i).y()
    if eff is not None:
        idx = np.argmax(eff)
        for c in cols:
            out += str(c[idx]) + '\t'
    else:
        print('Could not find column Eff')
    # averages
    for c in cols:
        out += str(np.median(c)) + '\t'
    out += '\n'
    # maybe save result in a file
    if isinstance(filename, str) and filesave is True:
        filesave = filename.replace('.txt', '_avgmax.txt')
    if filesave is not None:
        f = open(filesave, 'w')
        f.write(out)
        f.close()
    return out
    
    
    
    
    
def processSampleCellsMap(file, colorscale=None, figAx=None, pltClose=True, newGraphKwargs={}):
    newGraphKwargs = deepcopy(newGraphKwargs)
    newGraphKwargs.update({'silent': True})
    
    content = Graph(file, **newGraphKwargs)
    colToPlot = ['Voc', 'Jsc', 'FF', 'Eff', 'Rp', 'Rs', 'n', 'J0']
    inveScale = [False, False, False, False, False, True, True, True] # inverted color scale
    cols = content.getAttribute('collabels')
    rows = content.getAttribute('rowlabels')
    if not isinstance(cols, list):
        print('Error processSampleCellsMap: cols is not a list (value', cols, ')')
    if colorscale is None:
        if content.getAttribute('colorscale', None) is not None:
            colorscale = content.getAttribute('colorscale')
        if content.getAttribute('cmap', None) is not None:
            colorscale = content.getAttribute('cmap')
    filelist = []
    # combined plot
    graphs = [Graph('', **newGraphKwargs), Graph('', **newGraphKwargs)]
    axisheights = [[],[]]
    # main loop
    for i in range(len(colToPlot)):
        look = colToPlot[i]
        for j in range(len(cols)):
            c = cols[j]
            if c[:len(look)] == look and (len(c) <= len(look) or c[len(look)] in [' ','_','-','.','[',']','(',')']):
                c = c.replace('_', ' [').replace('(','[')
                if '[' in c and ']' not in c:
                    c += ']'
                c = c.replace('[pc]', '[%]').replace('mApcm2', 'mA/cm2')
                # sort somehow identical cells?
                vals = content.curve(j).y()
                filesave = '.'.join(file.split('.')[:-1]) + '_' + look
                filelist.append(filesave + '.txt')

                res = plotSampleCellsMap(rows, vals, c, colorscale=colorscale, filesave=filesave, figAx=figAx, inverseScale=inveScale[i], pltClose=pltClose)

                if isinstance(res, Graph) and res.length() > 0:
                    AB = 0 if i < 4 else 1
                    graphs[AB].append(Curve_Subplot([[0],[0]], {'subplotfile': res.getAttribute('filesave')+'.txt'}))
                    figsize = res.getAttribute('figsize', [6,4])
                    sadjust = res.getAttribute('subplots_adjust', [0.1,0.1,0.9,0.9])
                    graphs[AB].update({'figsize': figsize, 'subplots_adjust': sadjust})
                    axisheights[AB].append(figsize[1] * (sadjust[3]-sadjust[1]))
                break
    # set correct graph size for compiled graph
    bottom, top, hspace = 0.5, 0.5, 0.5
    totalh = [(sum(tmp) + (len(tmp)-1)*hspace + bottom + top) for tmp in axisheights]
    for i in range(2):
        figsize = graphs[i].getAttribute('figsize')
        sadjust = graphs[i].getAttribute('subplots_adjust')
        attr = {'figsize': [figsize[0], totalh[i]]}
        attr.update({'subplots_adjust': [sadjust[0], bottom/totalh[i], sadjust[2], 1-top/totalh[i], 0, hspace/axisheights[i][0]]})
        attr.update({'subplotsncols': 1, 'subplotsheight_ratios': axisheights[i]})
        graphs[i].update(attr)
        filesave = '.'.join(file.split('.')[:-1]) + '_' + ['basic','diode'][i]
        graphs[i].filename = filesave
        graphs[i].plot(filesave=filesave, figAx=figAx)
    return filelist
            
    

def plotSampleCellsMap(cells, values, title, colorscale=None, filesave='', figAx=None, inverseScale=False, pltClose=True, newGraphKwargs={}):
    sizeCell = np.array([0.6, 0.6])
    margin   = np.array([0.4, 0.4])

    newGraphKwargs = deepcopy(newGraphKwargs)
    newGraphKwargs.update({'silent': True})
    
    if len(values) == 0:
        return False

    # check cells are with correct form, i.e. 'a1'
    x, y, val = [], [], []
    split = [research(r'([a-zA-Z])([0-9]*)', c).groups() for c in cells]
    for i in range(len(split)):
        if len(split[i]) == 2:
            x.append(float(ord(split[i][0].lower())-96))
            y.append(float(split[i][1]))
            val.append(values[i]) # prefer to work on a copy and not modifying the list values
    x, y, val = np.array(x), np.array(y), np.array(val)
    if title == 'Voc [V]':
        title = 'Voc [mV]'
        val *= 1000
    title = title.replace('Voc_V','Voc [V]').replace('Jsc_mApcm2','Jsc [mA/cm2]').replace('FF_pc','FF [%]').replace('Eff_pc','Eff. [%]').replace('_Ohmcm2',' [Ohmcm2]')
    title = title.replace('A/cm2', 'A cm$^{-2}$').replace('J0 ','J$_0$ ').replace('Ohmcm2', '$\Omega$ cm$^2$')

    valNorm = val if not (val==0).all() and not len(val)==1 else [0.5]*len(val)
    if isinstance(colorscale, Colorscale):
        colorscale = colorscale.getColorScale()
    if isinstance(colorscale, str): # 'autumn' is a string, but a color list might not have been recognized as a list
        colorscale = stringToVariable(colorscale)
    if colorscale is None:
        colorscale = [[1,0,0], [1,1,0.5], [0,1,0]]
    if inverseScale and isinstance(colorscale, list):
        colorscale = colorscale[::-1]

    xticks = np.arange(0,max(x)+1,1) # if max(x) > 6 else np.arange(0,max(x)+1,1)
    yticks = np.arange(0,max(y)+1,1)
    axSize = np.array([sizeCell[0]*(max(xticks)-min(xticks)), sizeCell[1]*(max(yticks)-min(yticks))])
    figSize = axSize + 2 * margin[1]
    marg = margin / figSize
    txtCoords = np.transpose([(x - 0.5 - min(xticks)) / (max(xticks) - min(xticks)), (y - 0.5 - min(yticks)) / (max(yticks) - min(yticks))])
    toPrint = [roundSignificant(v, 3) for v in val]
    if np.average(val) > 1e2:
        toPrint = ['{:1.0f}'.format(v) for v in toPrint]
    if np.average(val) < 1e-3:
        toPrint = ['{:.1E}'.format(v).replace('E-0','E-') for v in toPrint]
    graph = Graph('', **newGraphKwargs)
    texttxt = []
    textarg = []
    for i in range(len(val)):
        texttxt.append(toPrint[i])
        textarg.append({'xytext': list(txtCoords[i]), 'xycoords': 'axes fraction', 'horizontalalignment':'center', 'verticalalignment':'center'})
    graph.append(Curve([x-0.5, y-0.5], {'type': 'scatter', 'marker': 's', 'markersize': (sizeCell[0]*72)**2, 'markeredgewidth': 0, 'cmap': colorscale}))#cmapParam
    graph.append(Curve([x-0.5, valNorm], {'type': 'scatter_c'}))
    
    graph.update({'subplots_adjust':[marg[0], marg[1], 1-marg[0], 1-marg[1]]})
    graph.update({'figSize': list(figSize)})
    graph.update({'text': texttxt, 'textargs': textarg})
    graph.update({'title': graph.formatAxisLabel(title)})
    graph.update({'xlim': [min(xticks), max(xticks)], 'ylim': [min(yticks), max(yticks)]})
    fct = []
    fct.append(['set_xticks', [list(xticks)], {}])
    fct.append(['set_yticks', [list(yticks)], {}])
    fct.append(['set_xticklabels', [[]], {}])
    fct.append(['set_yticklabels', [[]], {}])
    fct.append(['set_xticks', [list(xticks[1:]-0.5)], {'minor':True}])
    fct.append(['set_yticks', [list(yticks[1:]-0.5)], {'minor':True}])
    fct.append(['set_xticklabels', [[chr(int(i)-1+ord('a')) for i in xticks[1:]]], {'minor': True}])
    fct.append(['set_yticklabels', [[    int(i)             for i in yticks[1:]]], {'minor': True}])
    fct.append(['tick_params', [], {'axis': 'both', 'which': 'minor', 'length': 0}])
    fct.append(['grid', [True], {}])
    graph.update({'arbitraryfunctions': fct})

    if filesave is not None:
        graph.headers.update({'filesave': os.path.basename(filesave)})
        graph.plot(figAx=figAx, filesave=filesave)
    else:
        graph.plot(figAx=figAx, ifSave=False, ifExport=False, ifSubPlot=True)
    if pltClose and figAx is None:
        plt.close()
    return graph


	

if __name__ == "__main__":
    # go through files, store files content in order to later select pairs
    folder = './../examples/JV/SAMPLE_A/'
    processJVfolder(folder, fitDiodeWeight=5, pltClose=False)
#    processJVfolder(folder, groupCell=True, fitDiodeWeight=5, pltClose=False)


    file = r'./../examples/JV\SAMPLE_B_3layerMo\export_sample_b_3layermo_summary_illum.txt'
#    processSampleCellsMap(file, pltClose=False)
#    writeFileAvgMax(file)




# -*- coding: utf-8 -*-
"""
Created on Wed Feb 22 07:34:12 2017

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import numpy as np
import warnings
from copy import deepcopy


from grapa.graph import Graph
from grapa.graphIO import GraphIO
from grapa.curve import Curve
from grapa.mathModule import roundSignificant, derivative



class GraphCV(Graph):

    FILEIO_GRAPHTYPE = 'C-V curve'

    AXISLABELS = [['Voltage', 'V', 'V'], ['Capacitance', 'C', 'nF']] 

    @classmethod
    def isFileReadable(cls, fileName, fileExt, line1='', line2='', line3='', **kwargs):
        if fileExt == '.txt' and line1.strip()[:12] == 'Sample name:' and fileName[0:4] == 'C-V_':
           return True
        return False
    
        
    def readDataFromFile(self, attributes, **kwargs):
        len0 = self.length()
        GraphIO.readDataFromFileGeneric(self, attributes)
        self.castCurve('Curve CV', len0, silentSuccess=True)
        # label based on file name, maybe want to base it on file content
        self.curve(len0).update({'label': self.curve(len0).getAttribute('label').replace('C-V ','').replace(' [nF]','').replace(' Capacitance', '').replace('T=','')})
        #set [nF] units
        self.curve(len0).setY(1e9 * self.curve(len0).y())
        # compute phase angle if required
        nb_add = 0
        if self.getAttribute('_CVLoadPhase', False) is not False:
            f = 1e5 # [Hz]
            C = self.curve(len0).y() * 1e-9 # in [F], not [nF]
            conductance = None
            for c in range(len0+1, self.length()): # retrieve resistance curve, not always same place
                lbl = self.curve(c).getAttribute('Voltage')
                if isinstance(lbl, str) and lbl[:1] == 'R':
                    conductance = 1 / self.curve(c).y()
                    break
            if conductance is None:
                print('ERROR Read', self.filename, 'as CurveCf with phase: cannot find R.')
            else:
                phase_angle = np.arctan(f * 2 * np.pi * C / conductance) * 180. / np.pi
#                phase_angle = np.arctan(C / conductance) * 180. / np.pi
                self.append(Curve([self.curve(len0).x(), phase_angle], self.curve(len0).getAttributes()))
                nb_add += 1
        # delete unneeded Curves
        for c in range(self.length()-1-nb_add, len0, -1):
            self.deleteCurve(c)
        # cosmetics
        axisLabels = deepcopy(GraphCV.AXISLABELS)
        # normalize with area C
        area = self.curve(len0).getAttribute('cell area (cm2)', None)
        if area is None:
            area = self.curve(len0).getAttribute('cell area', None)
        if area is None:
            area = self.curve(len0).getAttribute('area', None)
        if area is not None:
            self.curve(len0).setY(self.curve(len0).y() / area)
            self.curve(len0).update({'cell area (cm2)': area})
            axisLabels[1][2] = axisLabels[1][2].replace('F', 'F cm$^{-2}$')
            if not self.silent:
                print('Capacitance normalized to area', self.curve(len0).getArea(),'cm2.')
        # graph cosmetics
        self.update({'xlabel': self.formatAxisLabel(axisLabels[0]),
                     'ylabel': self.formatAxisLabel(axisLabels[1])}) # default



class CurveCV(Curve):
    
    CURVE = 'Curve CV'

    CST_q = 1.6021766208e-19 # C # electrical charge
    CST_eps0 = 8.85418782e-12 # m-3 kg-1 s4 A2 # vacuum permittivity
    CST_epsR = 10 # relative permittivity
    
    CST_MottSchottky_Vlim_def = [0, 0.4]
    CST_MottSchottky_Vlim_adaptative = [-0.5, np.inf]
    
    
    def __init__(self, data, attributes, silent=False):
        # main constructor
        Curve.__init__(self, data, attributes, silent=silent)
        # for saved files further re-reading
        self.update ({'Curve': CurveCV.CURVE})
    


    # GUI RELATED FUNCTIONS
    def funcListGUI(self, **kwargs):
        out = Curve.funcListGUI(self, **kwargs)
        # format: [func, 'func label', ['input 1', 'input 2', 'input 3', ...]]
        out.append([self.setArea, 'Set cell area', ['New area'], [self.getArea()]])
        # fit Mott-Schottky
        if self.getAttribute('_fitFunc', None) is None or self.getAttribute('_popt', None) is None:
            out.append([self.CurveCV_fitVbiN, 'Fit Mott-Schotty', ['ROI [V]'], [CurveCV.CST_MottSchottky_Vlim_def]])
            out.append([self.CurveCV_fitVbiN_smart, 'Fit Mott-Schotty (around min N_CV)', ['Within range [V]'], [CurveCV.CST_MottSchottky_Vlim_adaptative]])
        else:
            param = roundSignificant(self.getAttribute('_popt'), 5)
            out.append([self.updateFitParam, 'Update fit', ['V_bi', 'N_CV'], [param[0], '{:1.4e}'.format(param[1])]])
        # set epsilon
        out.append([self.setEpsR, 'Set epsilon r', ['default = 10'], [self.getEpsR()]])
        out.append([self.printHelp, 'Help!', [], []])
# 1/C2 = 2 / q / Nc / Ks / eps0 / A2 (Vbi-V)
        
        return out
    
    def alterListGUI(self):
        out = Curve.alterListGUI(self)
        #out.append(['nm <-> eV', ['nmeV', ''], ''])
        out.append(['Mott-Schottky (1/C^2 vs V)', ['', 'CurveCV.y_ym2'], ''])
        out.append(['Carrier density N_CV [cm-3] vs V', ['', 'CurveCV.y_CV_Napparent'], ''])
        out.append(['Carrier density N_CV [cm-3] vs depth [nm]', ['CurveCV.x_CVdepth_nm', 'CurveCV.y_CV_Napparent'], ''])
        return out

    # FUNCTIONS used for curve transform (alter)
    def y_ym2(self, xyValue=None, **kwargs):
        """ Mott-Schottky plot: 1 / C**2 """
        if xyValue is not None:
            return xyValue[1]
        return 1 / (self.y(**kwargs) ** 2)
        
    def x_CVdepth_nm(self, **kwargs):
        """ apparent probing depth, assuming planar capacitor. """
        eps_r = CurveCV.getEpsR(self)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            w = eps_r * CurveCV.CST_eps0 / (1e-5 * self.y(**kwargs)) * 1e9 # y defined as nF/cm2, output depth in [nm]
        return w
    def y_CV_Napparent(self, xyValue=None, **kwargs):
        """ apparent carrier density N_CV """
        if xyValue is not None:
            return xyValue[1]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            dCm2dV = derivative(self.x(**kwargs), (1e-5 * self.y(**kwargs)) ** (-2)) # C from [nF cm2] to [F m-2]
        eps_r = CurveCV.getEpsR(self)
        N = - 2 / (CurveCV.CST_q * eps_r * CurveCV.CST_eps0 * (dCm2dV)) * 1e-6 # output unit in [cm-3]
        return N
    
    
    # FUNCTIONS RELATED TO GUI (fits, etc.)
    def setArea(self, value):
        oldArea = self.getArea()
        self.update({'cell area (cm2)': value})
        self.setY(self.y() / value * oldArea)
        return True
    def getArea(self):
        return self.getAttribute('cell area (cm2)', 1)
    
    # handle custom-defined eps_r
    def getEpsR(self):
        if self.getAttribute('epsR', 0) != 0:
            return self.getAttribute('epsR')
        return CurveCV.CST_epsR
    def setEpsR(self, value):
        self.update({'epsR': value})

        

    # functions for fitting Mott-Schottky plot
    def CurveCV_fitVbiN(self, Vlim=None, silent=False):
        """
        Returns a Curve based on a fit on Mott-Schottky plot.
        Calls fit_MottSchottkyto get parameters, then call func_MottSchottky
        to generate data.
        """
        Vbi, N_CV = self.fit_MottSchottky(Vlim=Vlim)
        attr = {'color': 'k', 'area': self.getArea(), '_popt': [Vbi, N_CV], '_fitFunc': 'func_MottSchottky', '_Vlim': Vlim, 'filename': 'fit to '+self.getAttribute('filename').split('/')[-1].split('\\')[-1]}
        if not silent:
            print('Fit Mott-Schottky plot: Vbi =', Vbi, 'V, apparent doping N_CV =', '{:1.4e}'.format(N_CV) + '.')
        return CurveCV([self.x(), self.func_MottSchottky(self.x(), Vbi, N_CV)], attr)
    def CurveCV_fitVbiN_smart(self, Vrange=None, window=[-2,2], silent=False):
        """
        Returns a Curve based on a fit on Mott-Schottky plot, after first 
        guessing best possible range for fifting (where N_CV is lowest)
        """
        Vlim = self.smartVlim_MottSchottky(Vlim=Vrange, window=window)
        return self.CurveCV_fitVbiN(Vlim=Vlim, silent=False)

    def fit_MottSchottky(self, Vlim=None):
        """
        Fits the C-V data on Mott-Schottky plot, in ROI Vlim[0] to Vlim[0].
        Returns built-in voltage Vbi [V] and apparent doping density N_CV [cm-3].
        """
        datax, datay = self.selectData(xlim=Vlim)
        datay =  1 / (datay * 1e-5) ** 2  # calculation in SI units [F m-2]
        z = np.polyfit(datax, datay, 1, full=True)[0]
        Vbi = - z[1] / z[0]
        N_CV = - 2 / (CurveCV.CST_q * CurveCV.getEpsR(self) * CurveCV.CST_eps0) / z[0]
        return Vbi, N_CV * 1e-6 # N_CV in [cm-3]
    def func_MottSchottky(self, V, Vbi, N_CV):
        """
        Returns C(V) which will appear linear on a Mott-Schottky plot.
        V_bi built-in voltage [V]
        N_CV apparent doping density [cm-3]
        Output: C [nF cm-2]
        """
        out = V * np.nan
        mask = (V < Vbi)
        Cm2 = 2 / (CurveCV.CST_q * CurveCV.getEpsR(self) * CurveCV.CST_eps0 * (N_CV * 1e6)) * (Vbi - V)
        out[mask] = (Cm2[mask] ** (-0.5))
        return out * 1e5 # output unit [nF is cm-2]
        
        
    def smartVlim_MottSchottky(self, Vlim=None, window=[-2,2]):
        """
        Returns Vlim [Vmin, Vmax] offering a possible best range for Mott-
        Schottky fit.
        Assumes V in monotoneous increasing/decreasing.
        Window: how many points around best location are taken. Default [-2,2]
        """
        V = self.x()
        N = self.y(alter='CurveCV.y_CV_Napparent')
        # identify index within given V limits
        if Vlim is None:
            Vlim = CurveCV.CST_MottSchottky_Vlim_adaptative
        Vlim = [min(Vlim), max(Vlim)]
        ROI = [np.inf, -np.inf]
        for i in range(len(V)):
            if V[i] >= Vlim[0]:
                ROI[0] = min(i, ROI[0])
            if V[i] <= Vlim[1]:
                ROI[1] = max(i, ROI[1])
        # identify best: smooth, take few points around minimum of N
        from scipy.signal import medfilt
        N_ = medfilt(N, 3) # thus we eliminate faulty points
        idx = np.argmin(N_[ROI[0]:ROI[1]])
        #print(self.getAttribute('temperature [k]'), [V[ROI[0]+idx+window[0]], V[ROI[0]+idx+window[1]]])
        return [V[ROI[0]+idx+window[0]], V[ROI[0]+idx+window[1]]]
        
        
        
    # other functions
    

    def printHelp(self):
        print('*** *** ***')
        print('CurveCV offer basic treatment of C-V curves of solar cells.')
        print('Input units must be [V] and [nF] (or [nF cm-2]).')
        print('Curve transforms:')
        print(' - Linear: standard is Capacitance per area [nF cm-2] vs [V].')
        print(' - Mott-Schottky: [1/(C cm-2)^2] vs V. Allows extraction of built-in voltage V_bi')
        print('   (intercept with y=0) and carrier density (from slope).')
        print(' - Carrier density N_CV [cm-3] vs V: the apparent carrier density is calculated')
        print('   from formula N_CV = -2 / (eps0 epsr d/dV(C^-2)).')
        print(' - Carrier density N_CV [cm-3] vs depth [nm]: the apparent depth is calculated from')
        print('   parallel plate capacitor C = eps0 epsr / d.')
        print('Analysis functions:')
        print(' - Set area: can normalize input data. For proper data analysis the units should be [nF cm-2].')
        print(' - Extract N_CV, Vbi: fits the linear segment on the Mott-Schottky plot.')
        print('   Select the suitable ROI before fitting (enter min and max Voltages).')
        return True
    



# -*- coding: utf-8 -*-
"""
@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import numpy as np
import copy
from scipy.optimize import fsolve


from grapa.graph import Graph
from grapa.graphIO import GraphIO
from grapa.datatypes.curveSIMS import CurveSIMS

from grapa.mathModule import roundSignificant, is_number


class GraphSIMS(Graph):
    """
    msmtId is an id based on the initial datafile name, in order to be able to
    process simultaneously several SIMS measurements
    class must be written like instance of self is actually a Graph object,
    not a GraphSIMS. Cannot call a method of GraphSIMS as method of self.
    """

    KEYWORDS = {'ggi': [['^71Ga+'],['^71Ga+', '^113In+']],
                'cgi': [['Cu+'],['^71Ga+', '^113In+']],
                'gi': [['^71Ga+', '^113In+'],[]],
                'cusn': [['Cu+'],['Sn+']],
                'cuzn': [['Cu+'],['Zn+']],
                'znsn': [['Zn+'],['Sn+']],
                'cuznsn': [['Cu+'],['Zn+', 'Sn+']],
                'cuge': [['Cu+'], ['^70Ge+']],
                'snge': [['Sn+'], ['^70Ge+']],
                }

    FILEIO_GRAPHTYPE = 'SIMS data'
    
    AXISLABELS = [['Sputter time', 't', 's'], ['Intensity', '', 'counts']]

    @classmethod
    def isFileReadable(cls, fileName, fileExt, line1='', line3='', **kwargs):
        if (fileExt == '.txt'  and
            (line1[0:7] == '\ttotal\t' or line3[0:16] == 'Sputter Time (s)')):
            return True
        return False
    
        
    def readDataFromFile(self, attributes, **kwargs):
        # open using fileGeneric
        GraphIO.readDataFromFileGeneric(self, attributes)
        if self.length() == 0:
            GraphIO.readDataFromFileGeneric(self, attributes, ifReplaceCommaByPoint=True)
        self.update({'typeplot': 'semilogy'})
        ylabel = self.curve(-1).getAttribute('sputter time (s)')
        if ylabel == '':
            ylabel = GraphSIMS.AXISLABELS[1]
        if self.curve(-1).getAttribute('sputter time (s)') != '':
            self.update({'xlabel': self.formatAxisLabel(GraphSIMS.AXISLABELS[0])})
            self.update({'ylabel': self.formatAxisLabel(ylabel)})
        # set correct labels
        for c in range(self.length()):
            self.curve(c).update({'label': self.curve(c).getAttribute('total')})
            self.castCurve('CurveSIMS', c, silentSuccess=True)
        # prints default keywords
        print('SIMS available ratio keywords:', ', '.join(key for key in GraphSIMS.KEYWORDS))
        # set default SIMS relative yields
        msmtId = self.curve(-1).getAttribute('_SIMSmsmt')
        GraphSIMS.setYieldCoefs(self, msmtId, ifAuto=True)
        # add temporary Ga+In curve
        ok = GraphSIMS.appendReplaceCurveRatio(self, msmtId, 'gi', 'Ga+In')
        if ok:
            # find edges of Ga+In curve
            c = GraphSIMS.getCurve(self, msmtId, 'Ga+In', silent=True)
            ROI = c.findLayerBorders(returnIdx=False) if c is not None else self.curve(-1).x(index=[0, self.curve(-1).shape(1)])
            GraphSIMS.setLayerBoundaries(self, msmtId, ROI)
        # by default, hide some curves - add others, for kestertes?
        hidden = []
        toHide = ['F+', 'Mg+', 'Al+', '^41K+', 'Fe+', '^65Cu+', '^66Zn+', 'Ga+', 'Se+', '^94Mo+', 'In+']
        for h in toHide:
            c = GraphSIMS.getCurve(self, msmtId, h, silent=True)
            if c is not None:
                c.update({'linestyle': 'none'})
                hidden.append(h)
        if len(hidden) > 0:
            print('SIMS curves hidden automatically:', ', '.join(hidden)+'.')

                    
                    
    
    def getLayerBoundaries(self, msmtId):
        for c in range(self.length()):
            if self.curve(c).getAttribute('_SIMSmsmt') == msmtId:
                return self.curve(c).getAttribute('_SIMSLayerBoundaries')
        print('ERROR graphSIMS getLayerBoundaries cannot find curve with desired msmtId (',msmtId,')')
        return False
    
    def setLayerBoundaries(self, msmtId, ROI):
        for c in range(self.length()):
            if self.curve(c).getAttribute('_SIMSmsmt') == msmtId:
                self.curve(c).update({'_SIMSLayerBoundaries': [min(ROI), max(ROI)]})
        self.update({'axvline': [min(ROI), max(ROI)]})
        return GraphSIMS.getLayerBoundaries(self, msmtId)

    def setLayerBoundariesDepthParameters(self, msmtId, ROI, depth):
        ROI = GraphSIMS.setLayerBoundaries(self, msmtId, ROI)
        offset, mult = None, None
        for c in range(self.length()):
            if self.curve(c).getAttribute('_SIMSmsmt') == msmtId:
                if offset is None:
                    ROI = np.argmin(np.abs(self.curve(c).x() - ROI[0])), np.argmin(np.abs(self.curve(c).x() - ROI[1]))
                    x = self.curve(c).x(index=ROI)
                    offset = - x[0]
                    mult = depth / (x[1]-x[0])
                self.curve(c).setDepthParameters(offset, mult)
        if offset is None:
            print('ERROR GraphSIMS setLayerBoundariesDepthParameters cannot find curve with desired msmtId (',msmtId,')')
        return True
    def setLayerBoundariesDepthParametersGUI(self, depth, ROI, msmtId=None):
        return GraphSIMS.setLayerBoundariesDepthParameters(self, msmtId, ROI, depth)

        
                
        

    def getCurve(self, msmtId, element, silent=False, returnIdx=False):
        for c in range(self.length()):
            if self.curve(c).getAttribute('_SIMSelement') == element and self.curve(c).getAttribute('_SIMSmsmt') == msmtId:
                if returnIdx:
                    return c
                return self.curve(c)
        if not silent:
            print ('ERROR getCurve: cannot find curve associated with element', element, 'and id',msmtId,'.')
            return False
        return None


    # function setting custom or defaults SIMS yields
    def setYieldCoefs(self, msmtId, elementYield={}, ifAuto=False):
        if ifAuto:
            default = {'^71Ga+': 1, '^113In+': 5.5, 'Cu+': 300}
            default.update(elementYield)
            elementYield = default
        for key in elementYield:
            c = GraphSIMS.getCurve(self, msmtId, key, silent=True)
            if c is not None:
                c.update({'_SIMSYieldCoef': elementYield[key]})
                print('Graph SIMS setYields', key, elementYield[key])

    
    def appendReplaceCurveRatio(self, msmtId, ratioCurves, curveId, attributes={}, savgol=None):
        """
        savgol: 2-element list width and degree for Savistky-Golay smoothening
        """
        if savgol is not None: # savistky golay smoothening of the Curves
            from scipy.signal import savgol_filter
            savgol = list(savgol)
            if len(savgol) < 2 or not is_number(savgol[0]) or not is_number(savgol[1]):
                savgol = None
                print('Warning GraphSIMS.appendReplaceCurveRatio: savgol must be a 2-elements list of ints for savistky-golay width and degree. Received', savgol)
            if savgol is not None: # tests on savlues
                savgol[0] = max(1, int((savgol[0]-1)/2) * 2 + 1, 1) # must be an odd number
                savgol[1] = min(max(int(savgol[1]), 1), savgol[0]-1)
            if savgol == [1,1]:
                savgol = None
            if savgol is not None:
                print('GraphSIMS: smooth curve data using Savistky-Golay window', savgol[0], 'degree', savgol[1])
        ratioCurves = GraphSIMS.getRatioProcessRatiocurves(self, ratioCurves)
        ret = True
        ratio = None
        x = None
        filename, layerBoundaries, depthOffset, depthMult = '', '', '', ''
        for i in range(len(ratioCurves)):
            for j in range(len(ratioCurves[i])):
                curve = GraphSIMS.getCurve(self, msmtId, ratioCurves[i][j], silent=True)
                if curve is None:
                    print('WARNING SIMS appendReplaceCurveRatio Curve not found: ', ratioCurves[i][j],'.')
                    ret = False
                    return False
                if ratio is None:
                    tmp = copy.deepcopy(curve.y())
                    ratio = np.array([tmp * 0.0, tmp * 0.0])
                try:
                    tmp = curve.y() * curve.getAttribute('_SIMSYieldCoef', default=1)
                    if savgol is not None:
                        tmp = savgol_filter(tmp, *savgol, deriv=0)
                    ratio[i,:] += tmp
                except ValueError as e:
                    print('ERROR GraphSIMS appendReplaceCurveRatio: curve not same size.')
                    print(ratioCurves[i][j], 'shape curve.y', curve.y().shape, 'into array', ratio[i,:].shape)
                    print(e)
                    print(curve.y())
                if x is None:
                    x = curve.x()
                if filename == '':
                    filename = curve.getAttribute('filename')
                if layerBoundaries == '':
                    layerBoundaries = curve.getAttribute('_SIMSlayerBoundaries')
                if depthOffset == '':
                    depthOffset = curve.getAttribute('_simsdepth_offset')
                if depthMult == '':
                    depthMult = curve.getAttribute('_simsdepth_mult')
        for i in range(len(ratioCurves)):
            if len(ratioCurves[i]) == 0:
                ratio[i,:] = 1
        attr = {'label': curveId, 'filename': filename, '_simslayerboundaries': layerBoundaries, '_simsdepth_offset': depthOffset, '_simsdepth_mult': depthMult}
        attr.update(attributes)
        new  = CurveSIMS([x, ratio[0]/ratio[1]], attributes=attr)
        c = GraphSIMS.getCurve(self, msmtId, curveId, silent=True, returnIdx=True)
        if c is None:
            self.append(new)
        else:
            self.replaceCurve(new, c)
        return ret
    def appendReplaceCurveRatioGUI(self, ratioCurves, curveId, msmtId=None):
        return GraphSIMS.appendReplaceCurveRatio(self, msmtId, ratioCurves, curveId)
    def appendReplaceCurveRatioGUISmt(self, ratioCurves, curveId, SGw, SGd, msmtId=None):
        return GraphSIMS.appendReplaceCurveRatio(self, msmtId, ratioCurves, curveId, savgol=[SGw, SGd])
        
 


    def getRatioProcessRatiocurves(self, ratioCurves):
        # handle keywords for ratiocurves
        if isinstance(ratioCurves, str):
            ratioCurves=ratioCurves.lower()
            if ratioCurves not in GraphSIMS.KEYWORDS:
                print('WARNING: SIMS setYieldsFromExternalData unknown key (',ratioCurves,'). Used GGI instead.')
                ratioCurves = 'ggi'
            for key in GraphSIMS.KEYWORDS:
                if ratioCurves == key:
                    ratioCurves = GraphSIMS.KEYWORDS[key]
                    print('Replaced keyword', key, 'with ratio', str(GraphSIMS.KEYWORDS[key])+'.')
                    break
        return ratioCurves
    
    def getRatio(self, msmtId, ROI, ratioCurves, ifReturnAvg=False, ifReturnCoefs=False):
        # output: [ratioOfAveragesOverROI, AverageOverROIofLocalRatio]
        ratioCurves = GraphSIMS.getRatioProcessRatiocurves(self, ratioCurves)
        # compute ratio
        avgOfRatio = np.array([ROI * 0.0, ROI * 0.0])
        ratioOfAvg = np.array([0.0, 0.0])
        retAvg = np.array([[0.0]*len(r) for r in ratioCurves])
        retCoefs = np.array([[1.0]*len(r) for r in ratioCurves])
        for i in range(len(ratioCurves)):
            for j in range(len(ratioCurves[i])):
                key = ratioCurves[i][j]
                curve = GraphSIMS.getCurve(self, msmtId, key, silent=True)
                if curve is not None:
                    retCoefs[i][j] = curve.getAttribute('_SIMSYieldCoef', default=1)
                    avgOfRatio[i,:] += (curve.y(index=ROI) * retCoefs[i][j])
                    retAvg[i][j] = np.average((curve.y(index=ROI))) * retCoefs[i][j]
                    ratioOfAvg[i] += retAvg[i][j]
                else:
                    print('ERROR: SIMS setYieldsFromExternalData cannot find curve',key,'.')
            if len(ratioCurves[i]) == 0:
                avgOfRatio[i,:] = ROI * 0.0 + 1
                ratioOfAvg[i] = 1
        RatioOfAvg = ratioOfAvg[0] / ratioOfAvg[1]
        AvgOfRatio = np.average(avgOfRatio[0][avgOfRatio[1]!=0] / avgOfRatio[1][avgOfRatio[1]!=0])
        out = [RatioOfAvg, AvgOfRatio, retAvg] if ifReturnAvg else [RatioOfAvg, AvgOfRatio]
        if ifReturnCoefs:
            out = out + [retCoefs]
        return out
    
    def targetRatioSetYield(self, msmtId, ROI, ratioCurves, element, target, silent=False):
        ratioInit = ratioCurves
        ratioCurves = GraphSIMS.getRatioProcessRatiocurves(self, ratioCurves)
        ratios = GraphSIMS.getRatio(self, msmtId, ROI, ratioCurves, ifReturnAvg=True)
        tunable = np.array([[False]*len(r) for r in ratioCurves])
        for i in range(len(ratioCurves)):
            for j in range(len(ratioCurves[i])):
                if ratioCurves[i][j] == element:
                    tunable[i][j] = True
        def func(coef, avgs, tunable, target):
            out0 = np.sum([avgs[0][j]*coef if tunable[0][j] else avgs[0][j] for j in range(len(avgs[0]))])
            out1 = np.sum([avgs[1][j]*coef if tunable[1][j] else avgs[1][j] for j in range(len(avgs[1]))])
            return out0 / out1 - target
        res = fsolve(func, 1, args=(ratios[2], tunable, target))
        if isinstance(res, np.ndarray):
            res = res[0]
        curve = GraphSIMS.getCurve(self, msmtId, element, silent=True)
        if curve is None:
            return False
        old = curve.getAttribute('_SIMSYieldCoef')
        curve.update({'_SIMSYieldCoef': res*old})
        if not silent:
            print('SIMS adjust yield to reach', ratioInit,'ratio', target,': ', curve.getAttribute('_SIMSelement'), 'yield', curve.getAttribute('_SIMSYieldCoef'), '(old value',old,')')
        return GraphSIMS.getRatioGUI(self, ROI, ratioInit, msmtId=msmtId, ifROIisIdx=True)
    def targetRatioSetYieldGUI(self, ROI, ratioCurves, element, target, msmtId=None):
        ROI = GraphSIMS.ROI_GUI2idx(self, msmtId, ROI)
        return GraphSIMS.targetRatioSetYield(self, msmtId, ROI, ratioCurves, element, target)


    def getRatioGUI(self, ROI, ratioCurves, msmtId=None, ifROIisIdx=False):
        if msmtId is None:
            print('ERROR GraphSIMS getRatioGUI: msmtId was not provided.')
            return False
        ratioName = str(ratioCurves)
        # GUI function: ROI is assumed is be function of sputtering time
        if not ifROIisIdx:
            ROI = GraphSIMS.ROI_GUI2idx(self, msmtId, ROI)
        GGIs = GraphSIMS.getRatio(self, msmtId, ROI, ratioCurves, ifReturnCoefs=True)
        out = ratioName + ': ' + '{:1.4f}'.format(GGIs[0]) + ' (avg of local ratios: ' + '{:1.4f}'.format(GGIs[1]) + '). Coefficients: ' + str(roundSignificant(GGIs[2][0],4)) + ', ' + str(roundSignificant(GGIs[2][1],4)) + '.'
        return out

    def ROI_GUI2idx(self, msmtId, ROI):
        c = None
        for curve in range(self.length()):
            if self.curve(curve).getAttribute('_SIMSmsmt') == msmtId:
                c = curve
                break
        if c is None:
            print('GraphSIMS getRatioGUI cannot find curve with correct msmtId (',msmtId,').')
        if ROI == '':
            return np.arange(0, self.curve(c).shape(1)-1)
        return np.arange(np.argmin(np.abs(self.curve(c).x() - min(ROI))), np.argmin(np.abs(self.curve(c).x() - max(ROI))))
        
        




# -*- coding: utf-8 -*-
"""
Created on Sat Oct 28 12:52:52 2017

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import numpy as np
from grapa.graph import Graph
from grapa.datatypes.curveEQE import CurveEQE


class GraphEQE_CSU(Graph):
    
    FILEIO_GRAPHTYPE = 'EQE curve'
    
    AXISLABELS = [['Wavelength', '\lambda', 'nm'], ['Cell EQE', '', '%']]
    
    @classmethod
    def isFileReadable(cls, fileName, fileExt, line1='', line3='', **kwargs):
        if (fileExt == '.txt' and line1 == 'CSU' and
            line3.startswith('Bias Light')):
            return True
        return False

    def readDataFromFile(self, attributes, **kwargs):
        attrs = {}
        # interpret headers
        f = open(self.filename, 'r')
        attrs['acquisition location'] = f.readline().strip()
        attrs['sample'] = f.readline().strip()
        attrs['label'] = attrs['sample']
        attrs['acquisition bias light'] = f.readline().replace('Bias Light ','').strip()
        attrs['acquisition cell area'] = float(f.readline().strip())
        attrs['acquisition Eg'] = float(f.readline().strip().replace('Eg = ',''))
        f.readline() # do nothing with it
        nlines = int(f.readline().strip())
        attrs['acquisition Jsc'] = float(f.readline().strip())
        attrs['collabels'] = '[' + ', '.join(f.readline().strip().split('\t')) + ']'
        f.close()
        
        # read data
        data = np.array([np.nan, np.nan])
        kw = {'delimiter': '\t', 'invalid_raise': False,
              'skip_header': 9, 'usecols': [0, 1]}
        try:
            data = np.transpose(np.genfromtxt(self.filename, **kw))
        except Exception as e:
            if not self.silent:
                print('WARNING GraphEQE_CSU cannot read file', self.filename)
                print(type(e), e)
        
        # create data
        self.append(CurveEQE(data, attributes))
        self.curve(-1).update(attrs) # file content may override default attributes
        if nlines != len(self.curve(-1).x()):
            print('WARNING GraphEQE_CSU: number of lines differ from expected')
        
        # cosmetics
        if len(data.shape) > 1:
            self.curve(-1).update({'mulOffset': 100})
        # some default settings
        self.update({'ylim': [0, 100], 'xlim': [300, np.nan],
                 'xlabel': self.formatAxisLabel(GraphEQE_CSU.AXISLABELS[0]),
                 'ylabel': self.formatAxisLabel(GraphEQE_CSU.AXISLABELS[1])})




# -*- coding: utf-8 -*-
"""
Created on Tue Feb  6 20:17:03 2018

@author: Romain
"""
import sys
import os

path = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..')
if path not in sys.path:
    sys.path.append(path)
from grapa.graph import Graph


graph = Graph('examples/PL/rem_Sample1_564_run2_i_TR.Sample.asc')
graph.plot()




# -*- coding: utf-8 -*-
"""
Created on Fri May  8 20:19:42 2015

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""
import numpy as np
import ast



def find_nearest(array,value):
    idx = (np.abs(array-value)).argmin()
    return array[idx]


def is_number(s):
    try:
        float(s)
        return True
    except Exception as e:
        return False

def roundSignificant(xSeries, nDigits):
    if isinstance(xSeries, (int, float)):
        out = roundSignificant([xSeries], nDigits)
        return out[0]
    try:
        out = [x if x==0 or np.isnan(x) or np.isinf(x) else np.round(x, int(nDigits-1-np.floor(np.log10(np.abs(x))))) for x in xSeries]
    except TypeError as e:
        print('TypeError in roundSignificant, returned input.')
        print(e)
        out = xSeries
    return np.array(out)


def roundgraphlim (lim):
    """ Returns rounded graph limits. lim=[1.5,3.0] for example. """
    lim = np.array(lim)
    lim.sort()
    diff = lim[1] - lim[0]
    target = np.array([lim[0] - diff/10, lim[1] + diff/10])
    magn = int(max(1 + np.ceil(-np.log10(np.abs(target)))))
#    print ('roundgraphlim', lim, target, magn)
    target[0] = np.floor(target[0] * 10**magn) / 10**magn
    target[1] = np.ceil (target[1] * 10**magn) / 10**magn
#    print ('   ',target)
    return target


def stringToVariable(val):
    try:
        val = float(val)
    except:
        try:
            val = ast.literal_eval(val)
#            if isinstance(val, list):
#                print ('   ', val[0], type(val[0]))
        except:
            if isinstance(val, str) and len(val) > 1:
                if (val[0] == '[' and val[-1] == ']') or (val[0] == '(' and val[-1] == ')'):
                    try :
                        val = [float(v) for v in val[1:-1].replace(' ','').replace('np.inf','inf').split(',')]
                    except:
                        pass
#                        try:
#                            val = [ast.literal_eval(v) for v in val[1:-1].replace(' ','').replace('np.inf','inf').split(',')]
#                        except:
#                            pass
    return val


def derivative(x, y):
    """ numerical derivative dy/dx for the x and y datapoint series.
    Symetrical differentialtion i+1, i+1
    x is supposed to be monotonic (sorted up/down) !
    """
    if not isinstance(x, (np.ndarray)):
        x = np.array(x)
    if not isinstance(y, (np.ndarray)):
        y = np.array(y)
    if len(x) != len(y):
        print ('ERROR Math derivative: x and y not same length!')
        print (x, y)
        return False
    if len(y) == 0:
        return np.array([])
    if len(y) == 1:
        return np.array([0])
    d = (y[1:]-y[:-1])/(x[1:]-x[:-1])
    return np.append(np.append([d[0]], 0.5*(d[:-1]+d[1:])), d[-1])

def derivative2nd(x,y):
    """ second numerical derivative d2y/dx2 for the x and y datapoint series.
    x is supposed to be monotonic (sorted up/down) !
    CAUTION: this function is locally more exact, but much more noisier than derivative(derivative)!
    Consider derivating twice on experiemental datasets.
    """
    if not isinstance(x, (np.ndarray)):
        x = np.array(x)
    if not isinstance(y, (np.ndarray)):
        y = np.array(y)
    if len(x) != len(y):
        print ('ERROR Math derivative: x and y not same length!')
        print (x, y)
        return False
    d = (y[1:]-y[:-1])/(x[1:]-x[:-1])
    dd= (d[1:]-d[:-1])/(x[2:]-x[:-2]) * 2
    return np.append(np.append((d[1]-d[0])/(x[0]-x[1]), dd), (d[-1]-d[-2])/(x[-1]-x[-2]))


def smooth(x,window_len=11,window='hanning'):
    """smooth the data using a window with requested size.
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.
    
    input:
        x: the input signal 
        window_len: the dimension of the smoothing window; should be an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal
        
    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)
    
    see also: 
    
    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter
 
    TODO: the window parameter could be the window itself if an array instead of a string
    NOTE: length(output) != length(input), to correct this: return y[(window_len/2-1):-(window_len/2)] instead of just y.
    """

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")
    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len<3:
        return x

    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")

    s = np.r_[x[window_len-1:0:-1],x,x[-1:-window_len:-1]]
    #print(len(s))
    if window == 'flat': #moving average
        w = np.ones(window_len,'d')
    else:
        w = eval('np.'+window+'(window_len)')

    y = np.convolve(w/w.sum(),s,mode='valid')
    return y[int((window_len-1)/2):int(len(y)-(window_len-1)/2)]
    
    


def xAtValue (xSeries, ySeries, value, xMinMax=[-np.inf, np.inf], silent=False) :
    x = np.nan
    # find x value at intercept 
    if len(xSeries) - len(ySeries) != 0 :
        print ('Error function xAtValues: lenght of xSeries and ySeries not identical!')
        print ('   len(xSeries): ', len(xSeries))
        print ('   len(ySeries): ', len(ySeries))
        return 0

    # should test if xSeries is sorted, otherwise should sort it (together with ySeries)
    # actually not needed
    idxBot = -1
    idxTop = -1
    limits = 2 - (xSeries > min(xMinMax)) - (xSeries < max(xMinMax))
    for i in range(len(xSeries)) :
        if limits[i] :
            continue;
        x =  xSeries[i]
        y =  ySeries[i]
#        print (i, x, y, idxBot, idxTop)
        if y <= value :
            if idxBot < 0 or np.abs(y - value) < np.abs(ySeries[idxBot] - value) :
                idxBot = i
                
        else :
            if idxTop < 0 or np.abs(y - value) < np.abs(ySeries[idxTop] - value) :
                idxTop = i

    if np.abs(idxBot - idxTop) != 1 :
        if not silent:
            print ('Error function xAtValues: possibly 2 crossing values, or data not sorted.', idxBot, idxTop, value)
            print (ySeries)

    if idxBot + idxTop == -2 :
        print ('Error function xAtValues: no suitable data found.')
    elif idxBot != -1 and idxTop != -1 :
        # linear interpolation
        x = xSeries[idxBot] + (xSeries[idxTop] - xSeries[idxBot]) * (value - ySeries[idxBot]) / (ySeries[idxTop] - ySeries[idxBot])
    elif idxBot != -1 :
        x = xSeries[idxBot]
    elif idxTop != -1 :
        x = xSeries[idxTop]
    if np.isnan(x):
        print ('ERROR mathModule xAtValue: cannot find suitable value.')
        print ('   xSeries', xSeries)
        print ('   ySeries', ySeries)
        print ('   value', value)
    return x
            
   


def bandgapFromTauc (nm, EQE, xLim=[600, 1500], yLim=[25, 70]) :
    if max(EQE) < 1 :
        print ('Function bandgapFromTauc: max(EQE) < 1. Multiplied by 100 for datapoint selection.')
        EQE *= 100
        
    mask = np.ones(len(nm), dtype=bool)
    for i in reversed(range (len (mask))) :
        if EQE[i] < yLim[0] or EQE[i] > yLim[1] or nm[i] < xLim[0] or nm[i] > xLim[1] :
            mask[i] = False
    nm = nm[mask]
    EQE = EQE[mask]

    eV = 1239.5 / nm
    tauc = (eV * EQE)**2
    if len (tauc > 1) :
        z = np.polyfit(eV, tauc, 1, full=True)[0]
        p = np.poly1d(z)
        bandgap = -z[1]/z[0]
        return [bandgap, z[0]]
    #print ('Function bandgapFromTauc: not enough suitable datapoints.')
    return [np.nan, np.nan]
   



def polynomFromDataSeries (xSeries, ySeries) :
    if len(xSeries) == 0 or len(ySeries) == 0 :
        print ('Error polynomFromDataSeries: cannot find polynom with no data.')
        print ('   length of series:',len(xSeries), len(ySeries))
    if len(xSeries) != len(ySeries) :
        print ('Error polynomFromDataSeries: cannot find polynom when length of x and y series do not match.')
        print ('   length of series:',len(xSeries), len(ySeries))
    degree = len(xSeries)
    print ('function not written yet')
    return -1







    



# -*- coding: utf-8 -*-
""" tk_ToolTip_class101.py
gives a Tkinter widget a tooltip as the mouse is above the widget
tested with Python27 and Python34  by  vegaseat  09sep2014
www.daniweb.com/programming/software-development/code/484591/a-tooltip-class-for-tkinter

Modified to include a delay time by Victor Zaccardo, 25mar16

Modified to allow user-defined wraplength, Romain Carron, 12nov2017
"""

try:
    # for Python2
    import Tkinter as tk
except ImportError:
    # for Python3
    import tkinter as tk

class CreateToolTip(object):
    """
    create a tooltip for a given widget
    """
    def __init__(self, widget, text='widget info', wraplength=180):
        self.waittime = 350     #miliseconds
        self.wraplength = wraplength   #pixels
        self.widget = widget
        self.text = text
        self.widget.bind("<Enter>", self.enter)
        self.widget.bind("<Leave>", self.leave)
        self.widget.bind("<ButtonPress>", self.leave)
        self.id = None
        self.tw = None

    def enter(self, event=None):
        self.schedule()

    def leave(self, event=None):
        self.unschedule()
        self.hidetip()

    def schedule(self):
        self.unschedule()
        self.id = self.widget.after(self.waittime, self.showtip)

    def unschedule(self):
        id = self.id
        self.id = None
        if id:
            self.widget.after_cancel(id)

    def showtip(self, event=None):
        x = y = 0
        x, y, cx, cy = self.widget.bbox("insert")
        x += self.widget.winfo_rootx() + 25
        y += self.widget.winfo_rooty() + 20
        # creates a toplevel window
        self.tw = tk.Toplevel(self.widget)
        # Leaves only the label and removes the app window
        self.tw.wm_overrideredirect(True)
        self.tw.wm_geometry("+%d+%d" % (x, y))
        label = tk.Label(self.tw, text=self.text, justify='left',
                       background="#ffffff", relief='solid', borderwidth=1,
                       wraplength = self.wraplength)
        label.pack(ipadx=1)

    def hidetip(self):
        tw = self.tw
        self.tw= None
        if tw:
            tw.destroy()

            
"""
# testing ...
if __name__ == '__main__':
    root = tk.Tk()
    btn1 = tk.Button(root, text="button 1")
    btn1.pack(padx=10, pady=5)
    button1_ttp = CreateToolTip(btn1, \
   'Neque porro quisquam est qui dolorem ipsum quia dolor sit amet, '
   'consectetur, adipisci velit. Neque porro quisquam est qui dolorem ipsum '
   'quia dolor sit amet, consectetur, adipisci velit. Neque porro quisquam '
   'est qui dolorem ipsum quia dolor sit amet, consectetur, adipisci velit.')

    btn2 = tk.Button(root, text="button 2")
    btn2.pack(padx=10, pady=5)
    button2_ttp = CreateToolTip(btn2, \
    "First thing's first, I'm the realest. Drop this and let the whole world "
    "feel it. And I'm still in the Murda Bizness. I could hold you down, like "
    "I'm givin' lessons in  physics. You should want a bad Vic like this.")
    root.mainloop()
"""


# -*- coding: utf-8 -*-
"""
Created on Wed Dec 27 10:38:53 2017

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import numpy as np
from grapa.curve import Curve

class CurveMycurve(Curve):

    # so the Curve class information is retrieved when opening an existing file
    CURVE = 'Curve mycurve'
    
    def __init__(self, data, attributes, silent=False):
        """
        Constructor with minimal structure: Curve.__init__, and then set the
        'Curve' parameter.
        """
        Curve.__init__(self, data, attributes, silent=silent)
        self.update ({'Curve': CurveMycurve.CURVE})


    # GUI RELATED FUNCTIONS
    def funcListGUI(self, **kwargs):
        """
        Fills in the Curve actions specific to the Curve type. Syntax:
        [func,
         'Button text',
         ['label 1', 'label 2', ...],
         ['value 1', 'value 2', ...],
         {'hidden': 'value'}, (optional)
         [dictFieldAttributes, {}, ...]] (optional)
        """
        out = Curve.funcListGUI(self, **kwargs)
        # already implemented easy access to Curve attributes
        at = ['label', 'color']
        out.append([self.updateValuesDictkeys, 'Set', at,
                    [self.getAttribute(a) for a in at], {'keys': at}])
        # some function
        choices =['min, max','number of points','just make a copy and add 0.1']
        out.append([self.statsAndAddition,
                    'Go',
                    ['Do some statistics'],
                    ['min, max'],
                    {},
                    [{'field':'Combobox', 'values': choices}]])
        out.append([self.printHelp, 'Help!', [], []])
        return out
    
    def alterListGUI(self):
        """
        Determines the possible curve visualisations. Syntax:
        ['GUI label', ['alter_x', 'alter_y'], 'semilogx']
        A few alter keywords are available but feel free to implement our own
        """
        out = Curve.alterListGUI(self)
        out.append(['derivative', ['', 'CurveMycurve.y_derivative'], ''])
        out.append(['log(abs(y))', ['', 'CurveMycurve.y_abs'], 'semilogy'])
        return out

    # data transform function
    def y_derivative(self, index=np.nan, xyValue=None):
        """
        Function computing the derivative on-the-fly during plotting.
        The same function is used to modify the graph xlim and ylim. In such
        cases the limits are passed through the xyValue.
        """
        # ylim cannot be computed from the initial limit values (derivative)
        if xyValue is not None:
            return np.nan
        # do not want to use self.x(index) syntax as we need more points to 
        # compute the local derivative. Here we compute over all data, then
        # restrict to the desired datapoint
        x, y = self.x(xyValue=xyValue), self.y(xyValue=xyValue)
        # compute derivative as symmetrical finite difference
        # for noisy data have a look to Savitsky-Golay or similar
        val = [(y[1]  - y[0]  ) / (x[1]  - x[0])]
        val+= list((y[2:] - y[:-2]) / (x[2:] - x[:-2]))
        val+= [(y[-2] - y[-1] ) / (x[-2] - x[-1])]
        # NB: This formula might not be valid if length is smaler than 2
        val = val[:len(x)]
        # restrict over required index datapoints
        if len(val) > 0:
            if np.isnan(index).any():
                return val[:]
            return val[index]
        return val
        
    # data transform function
    def y_abs(self, index=np.nan, xyValue=None):
        if xyValue is not None:
            # this way a negative ylim value can be ignored
            return np.abs([max(0.0, v) for v in xyValue[1]])
        return np.abs(self.y(index=index, xyValue=xyValue))
    
    # function called by user
    def statsAndAddition(self, what):
        if what == 'min, max':
            print('Values min', np.min(self.y()), ', max', np.max(self.y()))
        elif what == 'number of points':
            print('Number of points', len(self.y()))
        elif what == 'just make a copy and add 0.1':
            # return a copy of self after adding 1 to the y values
            out = self + 0.1 # basic operations are implemented and affect y
            out.update({'label': out.getAttribute('label')+' + 0.1'})
            return out
        else:
            print('Unknown keyword', what)
        return True
        
        
    def printHelp(self):
        print('*** *** ***')
        print('CurveMycurve is there just for show.')
        print('Maybe here some help for the user')
        return True



# -*- coding: utf-8 -*-
"""
Created on Fri Jul 15 15:46:13 2016

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import numpy as np


from grapa.graph import Graph
from grapa.graphIO import GraphIO
from grapa.curve import Curve
from grapa.mathModule import is_number, roundSignificant



class GraphJscVoc(Graph):
    
    FILEIO_GRAPHTYPE = 'Jsc-Voc curve'

    AXISLABELS = [['Voc', '', 'V'], ['Jsc', '', 'mA cm$^{-2}$']]
    
    @classmethod
    def isFileReadable(cls, fileName, fileExt, line1='', line2='', line3='', **kwargs):
        if fileExt == '.txt' and fileName[:7] == 'JscVoc_' and line1[:14] == 'Sample name: 	':
           return True # can open this stuff
        return False
    
    def readDataFromFile(self, attributes, **kwargs):
        le = self.length()
        GraphIO.readDataFromFileGeneric(self, attributes, **kwargs)
        # expect 3 columns data
        self.castCurve(CurveJscVoc.CURVE, le, silentSuccess=True)
        # remove strange characters from attributes keys
        attr = self.curve(le).getAttributes()
        dictUpd = {}
        for key in attr:
            if key.replace('²', '2').replace('Â²','2') != key:
                dictUpd.update({key.replace('²', '2').replace().replace('Â²','2'): attr[key]})
                dictUpd.update({key: ''})
        self.curve(le).update(dictUpd)
        self.curve(le).update({'type': 'scatter', 'cmap': [[0,0,1], [1,0.43,0]], 'markeredgewidth': 0})
        lbl = self.getAttribute('label').replace('JscVoc ','').replace('Â²','2').replace(' Values Jsc [mA/cm2]','')
        self.curve(le).update({'label': lbl})
        self.curve(le+1).update({'type': 'scatter_c'})
        #self.curve(le+1).swapShowHide() # hide T curve
        self.update({'typeplot': 'semilogy', 'alter': ['', 'abs'],
                     'xlabel': self.formatAxisLabel(GraphJscVoc.AXISLABELS[0]),
                     'ylabel': self.formatAxisLabel(GraphJscVoc.AXISLABELS[1])})

    
    def findCurveWithX(self, curve):
        """
        Find the Curve with same x data as the given Curve
        JscVoc: that Curve should store temperatures
        """
        for cu in range(self.length()):
            if self.curve(cu) == curve:
                # start looping at position of the curve
                for c in range(cu+1, self.length()):
                    if np.array_equiv(curve.x(), self.curve(c).x()):
                        return self.curve(c)
                # could not find, start looking from beginning
                for c in range(cu):
                    if np.array_equiv(curve.x(), self.curve(c)):
                        return self.curve(c)
                return False
    def splitTemperatures(self, curve, threshold=3):
        """
        Splits the compiled data into different data (one for each T)
        curve: stores the Jsc-Voc pairs - will need to find the T
        threshold in °C
        """
        T = GraphJscVoc.findCurveWithX(self, curve)
        if T == False:
            print('Error JscVoc: cannot find temperature Curve, must have same Voc list as Jsc-Voc Curve. Aborted.')
            return [], []
        Voc = curve.x()
        Jsc = curve.y()
        Tem = T.y()
        datas = []
        temps = []
        j = 0
        data, temp = [], []
        for i in range(len(Voc)):
            # check if is new temperature
            if j != 0:
                if np.abs(Tem[i] - np.average(temp)) > threshold:
                    datas.append(data)
                    temps.append(np.average(temp))
                    data = []
                    temp = []
                    j = 0
            if j == 0: # is new temperature
                data = [[Voc[i]], [Jsc[i]]]
                temp = [Tem[i]]
            else: # is not new
                data[0].append(Voc[i])
                data[1].append(Jsc[i])
                temp.append(Tem[i])
            j += 1
        datas.append(data)
        temps.append(np.average(temp))
        return datas, temps
    def CurvesJscVocSplitTemperature(self, threshold=3, curve=None):
        """
        Splits the compiled data into different data (one for each T)
        threshold in °C
        curve: stores the Jsc-Voc pairs. If None: error. Weird prototype design
            to allow calls from GUI
        """
        if curve is None:
            print('Error CurvesJscVocSplitTemperature, you must provide argument key "curve" with a Jsc-Voc Curve.')
        datas, temps = GraphJscVoc.splitTemperatures(self, curve, threshold=3)
        attr = curve.getAttributes()
        out = []
        for i in range(len(datas)):
            out.append(CurveJscVoc(datas[i], attr))
            out[-1].update({'temperature': temps[i], 'type': '', 'cmap': ''})
            out[-1].update({'label': out[-1].getAttribute('label')+' '+'{:.0f}'.format(temps[i])+'K'})
        return out
 
    # handling of curve fitting
    def CurveJscVoc_fitNJ0(self, Voclim=None, Jsclim=None, threshold=3, graphsnJ0=True, curve=None, silent=False):
        """
        Fit the Jsc-Voc data, returns fitted Curves.
        If required, first splits data in different temperatures.
        Voclim, Jsclim: fit limits for Voc and Jsc, in data units
        threshold in °C
        graphsnJ0: if True, also returns A vs T, J0 vs T and J0 vs A*T
        curve: stores the Jsc-Voc pairs. If None: error. Weird prototype design
            to allow calls from GUI
        """
        if curve is None:
            print('Error CurveJscVoc_fitNJ0, you must provide argument key "curve" with a Jsc-Voc Curve.')
            return False
        try:
            T = curve.T(default=np.nan, silent=True)
        except Exception:
            print('Warning CurveJscVoc_fitNJ0, method T() not found. Curve type not suitable?')
            T = np.nan
        # check if is single or multiple temperature
        if np.isnan(T):
            datas, temps = GraphJscVoc.splitTemperatures(self, curve, threshold=threshold)
        else:
            datas, temps = [[curve.x(), curve.y()]], [T]
        # fit different data series (different illuminations at same T)
        out = []
        ns, J0s = [], []
        for i in range(len(datas)):
            n, J0 = curve.fit_nJ0(Voclim=Voclim, Jsclim=Jsclim, data=datas[i], T=temps[i])
            attr = {'color': 'k', 'area': curve.getArea(), '_popt': [n, J0], 
                    '_fitFunc': 'func_nJ0', '_Voclim': Voclim, '_Jsclim': Jsclim,
                    'temperature': temps[i],
                    'filename': 'fit to '+curve.getAttribute('filename').split('/')[-1].split('\\')[-1]}
            if not silent:
                print('Fit Jsc-Voc (T =',temps[i],'): ideality factor n =', n, ', J0 =', '{:1.4e}'.format(J0), '[mA/cm2].')
            x, y = curve.selectData(xlim=Voclim, ylim=Jsclim, data=datas[i])
            out.append(CurveJscVoc([x, curve.func_nJ0(x, n, J0, T=temps[i])], attr))
            ns.append(n)
            J0s.append(J0)
        if graphsnJ0:
            from grapa.datatypes.curveArrhenius import CurveArrhenius, CurveArrheniusJscVocJ00
            # generate n vs T, J0 vs T
            lbl = curve.getAttribute('label')
            if len(lbl) > 0:
                lbl += ' '
            out.append(Curve([temps, ns],  {'linestyle': 'none', 'linespec': 'o', 'label': lbl+'Ideality factor A vs T [K]'}))
            out.append(Curve([temps, J0s], {'linestyle': 'none', 'linespec': 'o', 'label': lbl+'J0 vs T [K]'}))
            out.append(CurveArrhenius([np.array(temps)*np.array(ns), J0s], CurveArrheniusJscVocJ00.attr))
            out[-1].update({'label': lbl+out[-1].getAttribute('label')})# 'label': lbl+'J0 vs A*T'
            #out[-1].update({'type': 'scatter', 'markeredgewidth':0, 'markersize':100, 'cmap': 'inferno'})# 'label': lbl+'J0 vs A*T'
            #out.append(Curve([np.array(temps)*np.array(ns), np.array(temps)], {'linestyle': 'none', 'type': 'scatter_c'}))
        return out

            
        
class CurveJscVoc(Curve):

    CURVE = 'Curve JscVoc'
    
    q = 1.60217657E-19 # [C]
    k = 1.38E-23    # [J K-1]

    Tdefault = 273.15 + 25
    
    
    def __init__ (self, data, attributes, silent=False) :
        # delete area from attributes, to avoid normalization during initialization
        Curve.__init__ (self, data, attributes, silent=silent)
        self.update ({'Curve': CurveJscVoc.CURVE}) # for saved files further re-reading

    
    # RELATED TO GUI
    def funcListGUI(self, **kwargs):
        out = Curve.funcListGUI(self, **kwargs)
        # format: [func, 'func label', ['input 1', 'input 2', 'input 3', ...] (, [default1, default2, ...]) ]
        lbl = 'Area [cm2] (old value ' + "{:4.3f}".format(self.getArea()) + ')'
        out.append([self.setArea, 'Area correction', [lbl], [ "{:6.5f}".format(self.getArea())]]) # one line per function
        # fit function
        if self.getAttribute('_fitFunc') == '' or self.getAttribute('_popt', None) is None:
            out.append([GraphJscVoc.CurveJscVoc_fitNJ0, 'Fit J0 & ideality', ['Voclim', 'Jsclim', 'T max fluct.', 'compile'], [[0, max(self.x())], [0.1, max(self.y())], 3, True], {'curve': self}])
        else: # if fitted Curve
            out.append([self.updateFitParam, 'Modify fit', ['n', 'J0'], roundSignificant(self.getAttribute('_popt'), 5)])
        # split according to temperatures
        out.append([GraphJscVoc.CurvesJscVocSplitTemperature, 'Separate according to T', ['T max fluctuations'], [3], {'curve': self}])
        out.append([self.printHelp, 'Help!', [], []])
        return out
    
    def alterListGUI(self):
        out = Curve.alterListGUI(self)
        out.append(['Log10 abs', ['', 'abs'], 'semilogy'])
        return out


    # Handling of cell area
    def getArea(self):
        """ return area of the cell as stored in the Curve parameters """
        area = self.getAttribute('area [cm2]')
        if is_number(area):
            return area
        return 1
    def setArea(self, new):
        """ correct the cell area, and scale the y (list of Jsc) accordingly """
        old = self.getArea()
        self.setY(self.y() * old / new)
        self.update({'area [cm2]': new})
        return True
        
    def fit_nJ0(self, Voclim=None, Jsclim=None, data=None, T=None):
        """ perform fitting, returns best fit parameters """
        datax, datay = self.selectData(xlim=Voclim, ylim=Jsclim, data=data)
        if T is None:
            T = self.T()
        # actual fitting
        datay = np.log(datay)
        z = np.polyfit(datax, datay, 1, full=True)[0]
        n = self.q / self.k / T / z[0]
        J0 = np.exp(z[1])
        return [n, J0]
    def func_nJ0(self, Voc, n, J0, T=None):
        """ fit function """
        if T is None:
            T = self.T()
        out = J0 * np.exp(self.q * Voc / (n * self.k * T))
        return out
    
    def T(self, default=None, silent=False):
        """ Returns the acquisition temperature, otherwise default. """ 
        test = self.getAttribute('temperature', 0)
        if test != 0:
            return test
        if not silent:
            print('Curve JscVoc cannot find keyword temperature.', self.getAttributes())
        if default is None:
            return CurveJscVoc.Tdefault
        return default

    def printHelp(self):
        print('*** *** ***')
        print('CurveJV offer basic treatment of Jsc-Voc pairs of solar cells.')
        print('Based on Thomas Weiss script, and on the following references:')
        print('   Schock, Scheer, p111, footnote 20')
        print('   Hages et al., JAP 115, 234504 (2014)')
        print('Curve transforms:')
        print(' - Linear: standard is current density [mA cm-2] versus [V], at different light intensitites.')
        print(' - Log 10 abs: logarithm of J vs V. Same display as JV curve, to visualize the diode behavior.')
        print('Analysis functions:')
        print(' - Area correction: scale Jsc data according to cell area.')
        print(' - Fit J0 & ideality: fit Jsc vs Voc data and extract J0 and ideality factor A. Parameters:')
        print('   Voclim, Jsclim: fit limits for Voc and Jsc')
        print('   T max fluct.: identify groups of temperatures to fit only relevant data together.')
        print('       A new temperature is identified if a point deviates more than "value" from the average.')
        print('   compile: after fitting data, returns Curves with results: ideality factor versus T,')
        print('      J0 versus T, and J0 vs A * T.')
        print(' - Separate according to T: split data in several Curves according to the temperature identified.')
        print('      A new Curve is created once a point deviates more than "value" from the average.')
        return True




# -*- coding: utf-8 -*-
"""
Created on Tue Sep 13 15:44:58 2016

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import os
import numpy as np

from grapa.graph import Graph
from grapa.curve import Curve
from grapa.mathModule import is_number
    

# TODO: better parse perkin elmer Reflectnce/transmittance
# TODO: substract absorption in substrate


class CurveSpectrum(Curve):
    """
    Class handling optical spectra, with notably nm to eV conversion and
    background substraction.
    """
    
    CURVE = 'Curve Spectrum'
    FILE_INSTR_RESP = None
    FILE_SUBSTRATES = None
    LIST_SUBSTRATES = None
    
    def __init__(self, data, attributes, silent=False):
        # main constructor
        Curve.__init__(self, data, attributes, silent=silent)
        # initialize
        self.update ({'Curve': CurveSpectrum.CURVE})
        self.update({'_spectrumGUIoffset': True, '_spectrumGUIdark': True})
        subclass = self.getSubclass()
        if subclass is not None:
            subclass.initialize(self)
        

    # GUI RELATED FUNCTIONS
    def funcListGUI(self, **kwargs):
        out = Curve.funcListGUI(self, **kwargs)
        # format: [func, 'func label', ['input 1', 'input 2', 'input 3', ...]]
        # choice between subclass
        subclass = self.getSubclass()
        # nm <> eV conversion
        out.append([self.dataModifySwapNmEv, 'change data nm<->eV', [], []]) # one line per function
        # subtype
        choices = ['undefined']
        choices+= [s.__name__.replace('CurveSpectrum','') for s in self.subclasses]
        currentsubclass = subclass.__name__.replace('CurveSpectrum','') if subclass is not None else 'undefined'
        unit = str(self.getAttribute('_spectrumunit'))
        out.append([self.setSubclass, 'Save', ['Spectrum type', 'unit'],
                    [currentsubclass, unit], {},
                    [{'field': 'Combobox', 'values': choices}, 
                     {'field': 'Combobox', 'values': ['(no unit)', '%']}]])
        # offset
        if self.getAttribute('_spectrumGUIoffset'):
            out.append([self.addOffset, 'Add offset', ['new offset'], [self.getOffset()]]) # one line per function
        # dark
        if self.getAttribute('_spectrumGUIdark'):
            iddark = []
            if 'graph' in kwargs:
                for c in range(kwargs['graph'].length()):
                    iddark.append(str(c)+' '+kwargs['graph'].curve(c).getAttribute('label')[:6])
            else:
                print('graph must be provided to funcListGUI')
                kwargs['graph'] = None
            out.append([self.substractBG, 'Substract dark',
                        ['id dark', '', '', ''],
                        ['0', '1 interpolate both', '1 new curve', '0 ignore offset'],
                        {'graph': kwargs['graph'], 'operator': 'sub'},
                        [{'field':'Combobox', 'bind':'beforespace', 'width':4, 'values':iddark},
                         {'field':'Combobox', 'width':13,'values':['1 interpolate both', '2 interpolate dark', '0 element-wise']},
                         {'field':'Combobox', 'width':8, 'values':['1 new curve', '0 replace']},
                         {'field':'Combobox', 'width':13,'values':['0 ignore offsets', '1 with offsets']}]])
        # other, specific
        if subclass is not None and hasattr(subclass, 'funcListGUISpecific'):
            out += subclass.funcListGUISpecific(self, **kwargs)
        # help
        out.append([self.printHelp, 'Help!', [], []]) # one line per function
        return out
    
    def alterListGUI(self):
        out = Curve.alterListGUI(self)
        out += [['nm <-> eV', ['nmeV', ''], '']]
        out += [['nm <-> cm-1', ['nmcm-1', ''], '']]
        return out

    def getSubclass(self):
        self.subclasses = CurveSpectrum.__subclasses__()
        val = self.getAttribute('_spectrumSubclass')
        for s in self.subclasses:
            if val == s.__name__:
                return s
        return None
    def setSubclass(self, newtype, unit):
        """
        Spectrum type: set the Spectrum spectrum, affecting the GUI actions.
        Parameters:
        - newtype: subclass of CurveSpectrum. Examples:
          'CurveSpectrumReflectance', 'CurveSpectrumTransmittance', etc.
        - unit: empty, or '%' if data are in percent units.
        """
        # set unit
        if unit != '(no unit)':
            self.update({'_spectrumunit': unit})
        # set new type
        if newtype in ['', 'undefined']:
            self.update({'_spectrumSubclass': ''})
        else:
            if not newtype.startswith('CurveSpectrum'):
                newtype = 'CurveSpectrum' + newtype
            for s in CurveSpectrum.__subclasses__():
                if newtype == s.__name__:
                    self.update({'_spectrumSubclass': newtype})
                    return True
            print('CurveSpectrum.setSubclass: cannot find class', newtype)
            return False
        return True
    
    @classmethod
    def fileInstrResp(cls):
        if CurveSpectrum.FILE_INSTR_RESP is None:
            path = os.path.dirname(os.path.abspath(__file__))
            ref = Graph(os.path.join(path, 'spectrumInstrumentalResponses.txt'))
            CurveSpectrum.FILE_INSTR_RESP = ref
        return CurveSpectrum.FILE_INSTR_RESP
    @classmethod
    def fileSubstrates(cls):
        if CurveSpectrum.FILE_SUBSTRATES is None:
            path = os.path.dirname(os.path.abspath(__file__))
            ref = Graph(os.path.join(path, 'spectrumSubstrates.txt'))
            CurveSpectrum.FILE_SUBSTRATES = ref
        return CurveSpectrum.FILE_SUBSTRATES
    @classmethod
    def getListSubstrates(cls):
        """
        returns the substrates that have <substrate>+' R' and <substrate>+' T'
        in spectrumSubstrates.txt
        """
        if CurveSpectrum.LIST_SUBSTRATES is None:
            file = cls.fileSubstrates()
            lbls = [c.getAttribute('label') for c in file.iterCurves()]
            subsdict = {}
            for lbl in lbls:
                key = lbl[:-2]
                if key not in subsdict:
                    subsdict.update({key: ''})
                if lbl.endswith(' T'):
                    subsdict[key] = subsdict[key]+'T'
                if lbl.endswith(' R'):
                    subsdict[key] = subsdict[key]+'R'
            out = ['nope']
            for key in subsdict:
                if 'R' in subsdict[key] and 'T' in subsdict[key]:
                    out.append(key)
            CurveSpectrum.LIST_SUBSTRATES = out
        return CurveSpectrum.LIST_SUBSTRATES
   

        
    
    def getOffset(self):
        return self.getAttribute('_spectrumOffset', 0)
    def addOffset(self, value):
        """
        Add offset: adds a vertical offset to the data. The data are modified.
        The cumulated data correction is displayed, such that setting it to 0
        retrieves the original data (with some rounding errors)
        """
        if is_number(value):
            self.setY(self.y() + value - self.getOffset())
            self.update({'_spectrumOffset': value})
            return True
        return False

            
    # more "usual" methods
    def dataModifySwapNmEv(self):
        self.setX(self.NMTOEV / self.x())



    def substractBG(self, idDark, interpolate, ifnew, offsets, graph=None, **kwargs):
        """
        Substract dark: substract a curve to the data.
        Parameters:')
        - idDark: index of of the Curve containing the dark spectrum.
        - interpolate:
            0: performs element-wise substraction,
            1: output on x points of both data and dark, interpolate both,
            2: output on selected Curve x points, interpolate the dark Curve.
        - ifnew: 1: create a new Curve, 0: modify existing data.
        - offsets: 0: ignore offset and muloffset information.
            1: substract data after offset and muloffset operation.
        - graph: Graph object containing the dark Curve referred by idDark
        """
        # clean input from GUI: interpolate
        if interpolate in [0,1,2, '0','1','2']:
            interpolate = int(interpolate)
        elif isinstance(interpolate, str) and len(interpolate) > 0 and interpolate[0] in ['0','1','2']:
            interpolate = int(interpolate[0])
        else:
            interpolate = 1
        # clean input from GUI: ifNew
        def cleanInputBool(in_, default=False):
            if in_ in [0, 1, True, False, '0', '1', 'True', 'False']:
                if in_ in ['0', 'False']:
                    in_ = False
                else:
                    in_ = bool(in_)
            elif isinstance(in_, str) and len(in_) > 0 and in_[0] in ['0', '1']:
                in_ = True if in_[0] == '1' else False
            else:
                in_ = default
            return in_
        # clean input from GUI: ifNew, offsets
        ifnew = cleanInputBool(ifnew, default=True)
        offsets = cleanInputBool(offsets, default=False)
        idDark = min(int(idDark), graph.length()-1)
        j = None # index of 'curve' in graph
        for c in range(graph.length()):
            if graph.curve(c) == self:
                j = c
                break
        out = self.__add__(graph.curve(idDark), interpolate=interpolate, offsets=offsets, **kwargs)
        key = '_sub'
        msg = ('{Curve ' + (str(j) if j is not None else '') +
               ': ' + self.getAttribute('label') + '} - ' +
               '{Curve ' + str(idDark) + ': ' + graph.curve(idDark).getAttribute('label') + '}')
        out.update({key: msg})
        if not ifnew:
            if j is not None:
                graph.replaceCurve(out, j)
                print('CurveSpectrum.substractBG: modified Curve data.')
                return True
            print('CurveSpectrum.substractBG: cannot identify Curve!?! Created a new one instead.')
        print('CurveSpectrum.substractBG: created new Curve.')
        return out
    
    
    def correctInstrumentalResponse(self, instrresp, *args, **kwargs):
        """
        Correct for the instrumental response, using an instrumental response
        provided in file grapa/datatypes/spectrumInstrumentalResponses.txt
        !! Feel free to adapt the file with curves matching your system !!
        Parameters:
        - instrresp: label of the desired instr. resp. Curve in the file
        """
        file = self.fileInstrResp()
        for curve in file.iterCurves():
            if curve.getAttribute('label') == instrresp:
                out = self / curve
                val = self.getAttribute('_spectrumCorrected')
                out.update({'_spectrumCorrected': val + instrresp + ';',
                            'label': self.getAttribute('label')+' instr. resp. corr.'})
                self.update({'linestyle': 'none'})
                return out
        return 'Curve Spectrum instrumental response not found (required', instrresp, ', available', ', '.join([c.getAttribute('label') for c in file.iterCurves()]),')'
        
    def computeAbsorptance(self, auxcurve, graph=None, **kwargs):
        """
        Compute absorptance: computes an absorptance curve, defined as
            A% = 1 - R% - T%.
        Parameters:
        - auxcurve: index to a transmittance/reflectance curve,
        - graph: a graph containing the curve refered as auxcurve
        """
        out = self.computeAlpha(auxcurve, None, None, alpha=False, graph=graph,
                                **kwargs)
        out.update({'_spectrumalpha': '',
                    '_spectrumabsorptance': 'args auxcurve'+str(auxcurve)})
        return out

    
    def computeAlpha(self, auxcurve, thickness, substrate, alpha=True, graph=None, **kwargs):
        """
        Estimate alpha: estimates the optical absorption spectrum of a layer,
        using the formula
            alpha[cm-1] = - 1 / d[cm] * ln(T / (1-R)
        This formula is a reasonable approximation, assuming a few assumptions
        detailed in the manual. Notably, the substrate should be transparent, 
        and most of the reflections occurs at the air-layer interface. A
        substrate with low refractive index is to be preferred.
        Parameters:
        - auxcurve: index to a transmittance/reflectance curve,
        - thickness: thickness of the layer, in [nm],
        - substrate: a simple model can be used to account for the absorption
          in the substrate. The formula becomes:
            alpha[cm-1] = - 1 / d * ln(T / (1-R) * (1-Rsub) / Tsub).
        - graph: a graph containing the curve refered as auxcurve
        """
        if not isinstance(graph, Graph):
            print('Curve Spectrum computeAbsorptance, keyword "graph" not provided')
            return False
        if alpha:
            Rsub, Tsub = None, None
            if substrate not in ['', 'none']:
                filesubs = self.fileSubstrates()
                Rsub, Tsub = None, None
                for c in filesubs.iterCurves():
                    lbl = c.getAttribute('label')
                    if lbl == substrate + ' R':
                        Rsub = c
                    if lbl == substrate + ' T':
                        Tsub = c
                if Rsub is None or Tsub is None:
                    print('Curve Spectrum computeAbsorptance (alpha), cannot find substrate curves (',substrate,'), substrate ignored.')
                    Rsub, Tsub = None, None
        aux = None
        if isinstance(auxcurve, float):
            auxcurve = int(auxcurve)
            if auxcurve >= 0 and auxcurve < graph.length():
                aux = graph.curve(auxcurve)
        else:
            for c in range(graph.length()):
                if auxcurve == str(c)+' '+graph.curve(c).getAttribute('label'): 
                    aux = graph.curve(c)
                    break
        if aux is None:
            print('Curve Spectrum computeAbsorptance, auxiliary curve not found (',auxcurve,')')
            return False
        # determination R or T
        if (self.getAttribute('_spectrumSubclass') == 'CurveSpectrumTransmittance' or
             aux.getAttribute('_spectrumSubclass') == 'CurveSpectrumReflectance'):
            R, T = aux, self
        else:
            R, T = self, aux
        multR = 100 if R.getAttribute('_spectrumunit') == '%' else 1
        multT = 100 if T.getAttribute('_spectrumunit') == '%' else 1
        if alpha: # alpha
            out = T/multT / (1 - R/multR)
            if Rsub is not None and Tsub is not None:
                out *= (1 - Rsub) / Tsub
            out.setY(- 1 / (1e-7 * thickness) * np.log(out.y()))
            out.update({'_spectrumSubclass': ''})
        else: # absorptance
            out = 1 - R / multR - T / multT
            if multR == 100 and multT == 100:
                out.setY(out.y() * 100)
            out.update({'_spectrumSubclass': 'CurveSpectrumAbsorptance'})
        lbl = self.getAttribute('label').replace(' instr. resp. corr.','')
        replacelist = ['Reflectance', 'reflectance', 'R%', 'Transmittance', 'transmittance', 'T%']
        replacement = 'Absorptance' if not alpha else 'alpha [cm$^{-1}$]'
        flag = False
        for r in replacelist:
            if r in lbl:
                lbl = lbl.replace(r, replacement)
                flag = True
        if not flag:
            lbl = lbl + ' ' + replacement
        out.update({'label': lbl,
                    '_spectrumalpha': 'args auxcurve'+str(auxcurve)+', thickness'+str(thickness)+', substrate'+str(substrate)})
        return out
    
    def printHelp(self):
        print('*** *** ***')
        print('CurveSpectrum offers some capabilities to process optical spectrum data.')
        print('Curve transforms:')
        print(' - nm <-> eV: switch [eV] or [nm] data into the other representation (eV =', self.NMTOEV,'/ nm).')
        print('Analysis functions')
        print(' - Change data nm<->eV: changes data from nm to eV or inversely (eV =', self.NMTOEV,'/ nm),')
        self.printHelpFunc(CurveSpectrum.setSubclass)
        if self.getAttribute('_spectrumGUIoffset', True):
            self.printHelpFunc(CurveSpectrum.addOffset)
        if self.getAttribute('_spectrumGUIdark', True):
            self.printHelpFunc(CurveSpectrum.substractBG)
        sub = self.getSubclass()
        if sub is not None and hasattr(sub, 'printHelp_'):
            sub.printHelp_(self)
        return True



class CurveSpectrumReflectance(CurveSpectrum):
    def initialize(self):
        self.update({'_spectrumGUIoffset': False, '_spectrumGUIdark': False})
    def funcListGUISpecific(self, **kwargs):
        out = []
        file = self.fileInstrResp()
        graph = kwargs['graph'] if 'graph' in kwargs else None
        # instrumental response
        corr = '(already) ' if self.getAttribute('_spectrumCorrected') else ''
        lbls = [c.getAttribute('label') for c in file.iterCurves()]
        out.append([self.correctInstrumentalResponse,
                    'Correct',
                    [corr+'instrumental response'],
                    ['Reflectance (specular)'],
                    {},
                    [{'field': 'Combobox', 'width': 22, 'values': lbls}]])
        # compute absorptance, alpha
        if graph is not None:
            poss = []
            for c in range(graph.length()):
                if isinstance(graph.curve(c), CurveSpectrum):
                    if graph.curve(c).getAttribute('_spectrumSubclass') == 'CurveSpectrumTransmittance':
                        poss.append(str(c)+' '+graph.curve(c).getAttribute('label'))
            poss_ = poss[0] if len(poss) > 0 else ''
            out.append([self.computeAbsorptance, 'Compute absorptance',
                        ['select transmittance'],
                        [poss_],
                        {'graph': graph},
                        [{'field': 'Combobox', 'width': 20, 'values': poss}]])
            listsubs = self.getListSubstrates()
            out.append([self.computeAlpha, 'Estimate \u03B1',
                        ['transm.', 'thickness [nm]', 'substrate'],
                        [poss_, 50, 'none'],
                        {'graph': graph},
                        [{'field': 'Combobox', 'width': 12, 'values': poss},
                         {'width': 7},
                         {'field': 'Combobox', 'width': 6, 'values': listsubs}]])
        return out
    def printHelp_(self):
        print('CurveSpectrumReflectance offers 3 additional processing functions.')
        self.printHelpFunc(CurveSpectrum.correctInstrumentalResponse)
        self.printHelpFunc(CurveSpectrum.computeAbsorptance)
        self.printHelpFunc(CurveSpectrum.computeAlpha)
        

class CurveSpectrumTransmittance(CurveSpectrum):
    def initialize(self):
        self.update({'_spectrumGUIoffset': False, '_spectrumGUIdark': False})
    def funcListGUISpecific(self, **kwargs):
        out = []
        file = self.fileInstrResp()
        graph = kwargs['graph'] if 'graph' in kwargs else None
        # instrumental response
        corr = '(already) ' if self.getAttribute('_spectrumCorrected') else ''
        lbls = [c.getAttribute('label') for c in file.iterCurves()]
        value = 'Transmittance (specular)'
        if value not in lbls:
            if 'Transmittance (diffuse)' in lbls:
                value = 'Transmittance (diffuse)'
            if corr == '':
                corr = '(are you sure?) '
        out.append([self.correctInstrumentalResponse,
                    'Correct',
                    [corr+'instrumental response'],
                    [value],
                    {},
                    [{'field': 'Combobox', 'width': 22, 'values': lbls}]])
        # compute absorptance, alpha
        if graph is not None:
            poss = []
            for c in range(graph.length()):
                if isinstance(graph.curve(c), CurveSpectrum):
                    if graph.curve(c).getAttribute('_spectrumSubclass') == 'CurveSpectrumReflectance':
                        poss.append(str(c)+' '+graph.curve(c).getAttribute('label'))
            poss_ = poss[0] if len(poss) > 0 else ''
            out.append([self.computeAbsorptance, 'Compute absorptance',
                        ['select reflectance'],
                        [poss_ if len(poss) > 0 else ''],
                        {'graph': graph},
                        [{'field': 'Combobox', 'width': 20, 'values': poss}]])
            listsubs = self.getListSubstrates()
            out.append([self.computeAlpha, "Estimate \u03B1",
                        ['refl.', 'thickness [nm]', 'substrate'],
                        [poss_, 50, 'none'],
                        {'graph': graph},
                        [{'field': 'Combobox', 'width': 12, 'values': poss},
                         {'width': 7},
                         {'field': 'Combobox', 'width': 6, 'values': listsubs}]])
        return out
    def printHelp_(self):
        print('CurveSpectrumTransmittance offers 3 additional processing functions.')
        self.printHelpFunc(CurveSpectrum.correctInstrumentalResponse)
        self.printHelpFunc(CurveSpectrum.computeAbsorptance)
        self.printHelpFunc(CurveSpectrum.computeAlpha)

        
        
class CurveSpectrumAbsorptance(CurveSpectrum):
    def initialize(self):
        self.update({'_spectrumGUIoffset': False, '_spectrumGUIdark': False})
        

        



# -*- coding: utf-8 -*-
"""
Created on Sun Oct 29 14:19:52 2017

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

from grapa.graph import Graph
from grapa.curve import Curve

class Curve_Inset(Curve):
    """
    The purpose is this class is to provid GUI support to create insets
    """

    CURVE = 'inset'

    def __init__(self, *args, **kwargs):
        Curve.__init__(self, *args, **kwargs)
        # define default values for important parameters
        if self.getAttribute('insetfile') is '':
            self.update({'insetfile': ''})
        if self.getAttribute('insetcoords') is '':
            self.update({'insetcoords': [0.3,0.2,0.4,0.3]})
        if self.getAttribute('insetupdate') is '':
            self.update({'insetupdate': {}})
            
        if self.getAttribute('insetfile') is ' ':
            if self.getAttribute('subplotfile') not in ['', ' ']:
                self.update({'insetfile': self.getAttribute('subplotfile')})
                self.update({'subplotfile': ''})

        self.update ({'Curve': Curve_Inset.CURVE})

    # GUI RELATED FUNCTIONS
    def funcListGUI(self, **kwargs):
        out = Curve.funcListGUI(self, **kwargs)
        out.append([self.updateValuesDictkeys, 'Set', ['file inset'],   [self.getAttribute('insetfile')],   {'keys': ['insetfile']}])
        out.append([self.updateValuesDictkeys, 'Set', ['coords, in figure fraction [left, bottom, width, height]'],       [self.getAttribute('insetcoords')], {'keys': ['insetcoords']}])
        out.append([self.updateValuesDictkeys, 'Set', ['update inset'], [self.getAttribute('insetupdate')], {'keys': ['insetupdate']}])
        out.append([self.printHelp, 'Help!', [], []]) # one line per function
        return out

    
    
    def printHelp(self):
        print('*** *** ***')
        print('Class Curve_Inset facilitates the creation and customization of insets inside a Graph')
        print('Important parameters:')
        print('- insetfile: a path to a saved Graph, either absolute or relative to the main Graph.')
        print('  if set to ' ' (whitespace) or if no Curve are found in the given file, the next Curves of the main graph will be displayed in the inset.')
        print('- insetcoords: coordinates of the inset axis, relative to the main graph. Prototype: [left, bottom, width, height]')
        print('- insetupdate: a dict which will be applied to the inset graph. Provides basic support for run-time customization of the inset graph.')
        print("  Examples: {'fontsize':8}, or {'xlabel': 'An updated label'}")
        return True

    


# -*- coding: utf-8 -*-
"""
Created on Sat Nov 11 14:18:07 2017

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import tkinter as tk
from tkinter import ttk
from copy import deepcopy

import sys
import os
path = os.path.abspath(os.path.join('.', '..', '..'))
if path not in sys.path:
    sys.path.append(path)

from grapa.mathModule import stringToVariable
from grapa.gui.createToolTip import CreateToolTip
from grapa.gui.GUImisc import EntryVar, CheckbuttonVar, OptionMenuVar, ComboboxVar


class GuiManagerAnnotations(tk.Frame):
    """
    Provides a popup window for quick and easier modification of text
    annotations, legends, legend titles and graph titles.
    """
    # Implementation started with a not-great code for textx, then a more
    # tunable solution was designed for the rest. Previous not modified as it
    # was working
    
    def __init__(self, master, graph, funcupdate):
        self.master = master
        tk.Frame.__init__(self, master)
        self.funcupdate = funcupdate
        self.graph = graph
        # save intial values
        self.oldattr = {}
        self.oldattr['text']     = deepcopy(graph.getAttribute('text'))
        self.oldattr['textxy']   = deepcopy(graph.getAttribute('textxy', ['']))
        self.oldattr['textargs'] = deepcopy(graph.getAttribute('textargs', [{}]))
        self.oldattr['legendtitle'] = deepcopy(graph.getAttribute('legendtitle'))
        # prepare
        self.annotations = []
        self.configText()
        self.configLegTit()
        self.configLegend()
        # fill GUI
        self.frame = tk.Frame(self.master)
        self.initFonts(self.frame)
        
        self.fillUIMain(self.frame)
        self.frame.pack(side='top')
        self.master.bind("<Return>", lambda event: self.go())

    def initFonts(self, frame):
        import tkinter.font as font
        a = tk.Label(frame, text='')
        self.fontBold = font.Font(font=a['font'])
        self.fontBold.configure(weight='bold')
    

    def configText(self):
        self.textfields = []
        self.textfields.append({'label': 'visible','field': 'Checkbutton', 'fromkwargs': 'visible', 'tooltip': 'Visible?', 'default': True})
        self.textfields.append({'label': 'text',        'width': 15,  'tooltip': 'Enter a text'})
        self.textfields.append({'label': 'color',       'width': 7,  'fromkwargs': 'color',     'tooltip':'Color. Example: "r", "[1,0,1]", or "pink"', 'field': 'Combobox', 'values': ['r', 'pink', '[1,0,1]']})
        self.textfields.append({'label': 'fontsize',    'width': 7,  'fromkwargs': 'fontsize',  'tooltip':'Fontsize. Example: "16"'})
        self.textfields.append({'label': 'xytext',      'width': 12, 'fromkwargs': 'xytext',    'tooltip':'Position of the text. Example: "(0.2,0.3)"'})
        self.textfields.append({'label': 'textcoords', 'field': 'OptionMenu', 'fromkwargs': 'textcoords','tooltip':'By default "figure fraction". Can also be "axes fraction", "figure pixels", "data", etc.', 'values': ['figure fraction', 'figure pixels', 'figure points', 'axes fraction', 'axes pixels', 'axes points', 'data']})
        self.textfields.append({'label':'other properties','width': 38,
                                'tooltip':"kwargs to ax.annotate. Example: \"{'verticalalignment':'center', 'xy':(0.5,0.5), 'xycoords': 'figure fraction', 'rotation': 90, 'arrowprops':{'facecolor':'blue', 'shrink':0.05}}\"\nSee https://matplotlib.org/api/_as_gen/matplotlib.axes.Axes.annotate.html",
                                'default':{}, 'field': 'Combobox',
                                'values': ["{}",
                                           "{'verticalalignment':'center'}",
                                           "{'xy':(0.5,0.5), 'xycoords': 'figure fraction', 'arrowprops':{'facecolor':'b', 'shrink':0.05}}",
                                           "{'rotation': 90}"]})
        for field in self.textfields:
            if 'field' not in field:
                field['field'] = 'Entry'
            if 'width' not in field:
                field['width'] = -1
            if 'fromkwargs' not in field:
                field['fromkwargs'] = ''
        self.textvars = []
        for field in self.textfields:
            self.textvars.append([])
    def configLegTit(self):
        self.legtitfields = []
        self.legtitfields.append({'label': 'Legend title','width': 0})
        self.legtitfields.append({'label': 'title',       'width': 15,  'tooltip': 'Enter a title to the legend'})
        self.legtitfields.append({'label': 'color',       'width': 7,   'fromkwargs': 'color',    'tooltip':'Color. Example: "r", "[1,0,1]", or "pink"', 'field': 'Combobox', 'values': ['r', 'pink', '[1,0,1]']})
        self.legtitfields.append({'label': 'fontsize',    'width': 7,   'fromkwargs': 'fontsize', 'tooltip':'Numeric, eg. "20"'})
        self.legtitfields.append({'label': 'align', 'field': 'OptionMenu', 'fromkwargs': 'align',    'tooltip':'Horizontal alignment of title. eg. "left", "right".\nFor legendtitle: not a regular matplotlib keyword.', 'values': ['left', 'center', 'right']})
        self.legtitfields.append({'label': 'position (x,y)','width': 10,'fromkwargs': 'position',
                                  'tooltip':'legendtitle: (x,y) position shift in pixels, eg. "(20,0)"\ntitle: (x,y) relative position, eg. "(0,1)"', 'field': 'Combobox', 'values': ['(10,0)', '(-10,20)']})
        self.legtitfields.append({'label':'other properties','width': 40,'tooltip':'kwargs to legend.set_title()'})
        for field in self.legtitfields:
            if 'field' not in field:
                field['field'] = 'Entry'
            if 'width' not in field:
                field['width'] = -1
        self.legtitvars = []
        for field in self.legtitfields:
            self.legtitvars.append([])
    def configLegend(self):
        self.legendfields = []
        self.legendfields.append({'label': 'Legend',    'width': 0})
        self.legendfields.append({'label': 'color',     'width': 7, 'tooltip':'Color. Example: "r", "[1,0,1]", or "pink"\n"curve" will give same color as trace', 'fromkwargs': 'color', 'field': 'Combobox', 'values': ['r', 'pink', '[1,0,1]', 'curve']})
        self.legendfields.append({'label': 'fontsize',  'width': 7, 'tooltip':'Numeric, eg. "20"', 'fromkwargs': 'fontsize'})
        self.legendfields.append({'label': 'loc',  'field': 'OptionMenu', 'tooltip':'Location, eg. "nw" (north west), "se", "center"', 'fromkwargs': 'loc', 'default': 'best', 'values': ['best', 'nw', 'n', 'ne', 'w', 'center', 'e', 'sw', 's', 'se']})
        self.legendfields.append({'label': 'ncol',      'width': 6, 'tooltip':'Number of columns, eg. "2"', 'fromkwargs': 'ncol', 'casttype': int})
        self.legendfields.append({'label': 'bbox_to_anchor','width': 12, 'tooltip':'(x,y) shift to base position in axes fraction, eg. "(0.1, 0.2)"', 'fromkwargs': 'bbox_to_anchor', 'field': 'Combobox', 'values': ['(0,0.1)', '(0.1,0.2)']})
        self.legendfields.append({'label':'other properties','width': 40,
                                  'tooltip':'kwargs to ax.legend(), see https://matplotlib.org/api/_as_gen/matplotlib.axes.Axes.legend.html',
                                  'field': 'Combobox',
                                  'values': ["{}", "{'frameon': True, 'framealpha':1}", "{'framealpha': 1}", "{'numpoints': 2}"]})
        for field in self.legendfields:
            if 'field' not in field:
                field['field'] = 'Entry'
            if 'width' not in field:
                field['width'] = -1
        self.legendvars = []
        for field in self.legendfields:
            self.legendvars.append([])
        
    
    def parseGraph(self):
        # retrieve data from Graph object, ensure formatting is ok
        self.attr = {}
        self.attr['text']     = deepcopy(self.graph.getAttribute('text'))
        self.attr['textxy']   = deepcopy(self.graph.getAttribute('textxy', ['']))
        self.attr['textargs'] = deepcopy(self.graph.getAttribute('textargs', [{}]))
        if not isinstance(self.attr['text'], list):
            self.attr['text'] = [self.attr['text']]
        if not isinstance(self.attr['textxy'], list):
            self.attr['textxy'] = [self.attr['textxy']]
        if not isinstance(self.attr['textargs'], list):
            self.attr['textargs'] = [self.attr['textargs']]
        while len(self.attr['textxy']) < len(self.attr['text']):
            self.attr['textxy'].append('')
        while len(self.attr['textargs']) < len(self.attr['text']):
            self.attr['textargs'].append({})
        # override xytext in textargs with content of textxy - later can ignore textxy
        for i in range(len(self.attr['textxy'])):
            if self.attr['textxy'][i] != '':
                self.attr['textargs'].update({'xytext': self.attr['textxy'][i]})
        if self.attr['text'] == ['']:
            self.attr['text'] = []
        
    def fillUIMain(self, frame):
        self.frameGrid = tk.Frame(frame)
        self.frameGrid.pack(side='top', anchor='w')
        self.fillUITextGrid(self.frameGrid)
        
        self.legendGrid = tk.Frame(frame)
        self.legendGrid.pack(side='top', anchor='w')
        self.fillUILegendGrid(self.legendGrid)
        
        self.legTitGrid = tk.Frame(frame)
        self.legTitGrid.pack(side='top', anchor='w')
        self.fillUILegendTitleGrid(self.legTitGrid)
        
        frm = tk.Frame(frame)
        frm.pack()
        self.fillUIButtons(frm)
    
    def fillUIButtons(self, frame):
        self.ButtonRevert = tk.Button(frame, text='Revert to initial', fg='grey', command=self.revert)
        self.ButtonRevert.pack(side='left')
        self.ButtonQuit = tk.Button(frame, text='Quit', fg='grey', command=self.close_windows)
        self.ButtonQuit.pack(side='left', padx=10)
        self.ButtonGo = tk.Button(frame, text='Update graph', command=self.go, font=self.fontBold)
        self.ButtonGo.pack(side='right', padx=30)
    
    def fillUITextGrid(self, frame):
        self.parseGraph()
        # remove previsously created widgets, start afresh
        for row in self.annotations:
            for field in row:
                if hasattr(field, 'destroy'):
                    field.destroy()
        self.annotations = []
        args = []
        for j in range(len(self.attr['text'])):
            self.annotations.append([''] * len(self.textfields))
            args.append(deepcopy(self.attr['textargs'][j]))
        self.annotations.append([''] * len(self.textfields)) # for new text
        args.append({}) # defautl new
        for j in range(len(self.textfields)):
            field = self.textfields[j]
            e = tk.Label(frame, text=field['label'])
            e.grid(row=0, column=(j+1))
            if 'tooltip' in field:
                CreateToolTip(e, field['tooltip'])
            # data fields and new text fields
            kw = field['fromkwargs']
            for i in range(len(self.attr['text'])+1):
                value = ''
                if i == len(self.attr['text']): # "new" text field
                    value = ''
                elif field['label'] == 'text': # text: take it from attr['text']
                    value = self.attr['text'][i].replace('\n', '\\n')
                else: # otherwise must grab it from the textargs dict
                    if kw != '' and kw in args[i]:
                        value = args[i][kw]
                        del args[i][kw]
                    elif j == len(self.textfields) - 1:
                        value = str(args[i])
                if value == '' and 'default' in field:
                    value = field['default']
                e = None
                if field['field'] == 'OptionMenu':
                    e = OptionMenuVar(frame, values=field['values'], default=str(value))
                elif field['field'] == 'Checkbutton':
                    e = CheckbuttonVar(frame, '', default=(bool(value) if value != '' else False))
                if field['field'] == 'Combobox':
                    e = ComboboxVar(frame, field['values'], default=str(value), width=field['width'])
                elif field['width'] > 0:
                    e = EntryVar(frame, str(value), width=field['width'])
                if e is not None:
                    self.annotations[i][j] = e
                    e.grid(column=(j+1), row=(i+1), pady=0, ipady=0, padx=1)
        tk.Label(frame, text='Text', font=self.fontBold).grid(row=0, column=0)
        tk.Label(frame, text='New',  font=self.fontBold).grid(row=(len(self.attr['text'])+1), column=0)

    def fillUILegendGrid(self, frame):
        Labels = []
        for j in range(len(self.legendfields)):
            field = self.legendfields[j]
            Labels.append(tk.Label(frame, text=field['label']))
            if field['width'] == 0:
                Labels[-1].configure(font=self.fontBold)
            Labels[j].grid(row=(0 if field['width']!=0 else 1), column=j)
            if 'tooltip' in field:
                CreateToolTip(Labels[j], field['tooltip'])
            self.legendvars[j].append(tk.StringVar())
            e = None
            if field['field'] == 'OptionMenu':
                e = tk.OptionMenu(frame, self.legendvars[j][-1], *field['values'])
            if field['field'] == 'Combobox':
                e = ttk.Combobox(frame, textvariable=self.legendvars[j][-1], values=field['values'], width=field['width'])
            elif field['width'] > 0:
                e = tk.Entry(frame, width=field['width'], textvariable=self.legendvars[j][-1])
            if e is not None:
                e.grid(column=j, row=len(self.legendvars[j]), padx=1)
        self.legendTitleFillValues('legendproperties', self.legendfields, self.legendvars)
    def fillUILegendTitleGrid(self, frame):
        Labels = []
        # legend title
        for j in range(len(self.legtitfields)):
            field = self.legtitfields[j]
            if field['width'] != 0:
                Labels.append(tk.Label(frame, text=field['label']))
                Labels[-1].grid(row=0, column=j)
            if 'tooltip' in field:
                CreateToolTip(Labels[-1], field['tooltip'])
            self.legtitvars[j].append(tk.StringVar())
            e = None
            if field['field'] == 'OptionMenu':
                e = tk.OptionMenu(frame, self.legtitvars[j][-1], *field['values'])
            if field['field'] == 'Combobox':
                e = ttk.Combobox(frame, textvariable=self.legtitvars[j][-1], values=field['values'], width=field['width'])
            elif field['width'] > 0:
                e = tk.Entry(frame, width=field['width'], textvariable=self.legtitvars[j][-1])
            if e is not None:
                e.grid(column=j, row=len(self.legtitvars[j]), ipady=0, pady=0, padx=1)
        # graph title
        self.legtitfields[5].update({'values': ["(0,1)", "(0.5, 0.95)"]})
        for j in range(len(self.legtitfields)):
            field = self.legtitfields[j]
            # graph title
            self.legtitvars[j].append(tk.StringVar())
            if 'field' not in field:
                field['field'] = 'Entry'
            e = None
            if field['field'] == 'OptionMenu':
                e = tk.OptionMenu(frame, self.legtitvars[j][-1], *field['values'])
            if field['field'] == 'Combobox':
                e = ttk.Combobox(frame, textvariable=self.legtitvars[j][-1], values=field['values'], width=field['width'])
            elif field['width'] > 0:
                e = tk.Entry(frame, width=field['width'], textvariable=self.legtitvars[j][-1])
            if e is not None:
                e.grid(column=j, row=len(self.legtitvars[j]), ipady=0, pady=0, padx=1)
        tk.Label(frame, text='Legend title', font=self.fontBold).grid(row=1, column=0)
        tk.Label(frame, text='Graph title',  font=self.fontBold).grid(row=2, column=0, sticky='W')
        self.legendTitleFillValues('legendtitle', self.legtitfields, self.legtitvars, row=0)
        self.legendTitleFillValues('title', self.legtitfields, self.legtitvars, row=1)
            
        
    def legendTitleFillValues(self, attribute, fields, vars_, row=0):
        attr = deepcopy(self.graph.getAttribute(attribute))
        vals = [''] * len(fields)
        vals[-1] = attr
        if attribute in ['title', 'legendtitle']:
            if not isinstance(attr, list):
                attr = [attr, {}]
            vals[1] = attr[0]
            vals[-1] = attr[1]
        elif attribute in ['legendproperties']:
            if vals[-1] == '':
                vals[-1] = {}
        for i in range(len(fields)):
            if 'fromkwargs' in fields[i]:
                key = fields[i]['fromkwargs']
                if attribute == 'title' and key == 'align':
                    key = 'loc' # different keyword for same stuff
                if key in vals[-1]:
                    vals[i] = str(vals[-1][key])
                    del vals[-1][key]
            if vals[i] == '' and 'default' in fields[i]:
                vals[i] = fields[i]['default']
        vals[1] = vals[1].replace('\n','\\n')
        for j in range(0, len(vars_)):
            vars_[j][row].set(vals[j])

    
    def go(self):
        # text annotations
        out = {'text': [], 'textxy': [], 'textargs': []}
        i = 0
        for ann in self.annotations:
            if ann[1].get() is not '':
                out['text'].append(ann[1].get().replace('\\n', '\n'))
                args = stringToVariable(ann[-1].get())
                if not isinstance(args, dict):
                    print('GuiManagerAnnotations invalid input:', args, 'should be a dict')
                    args = {}
                for j in range(len(self.textfields)-1):
                    if self.textfields[j]['fromkwargs'] != '':
                        val = stringToVariable(ann[j].get())
                        if self.textfields[j]['label'] == 'xytext':
                            if not isinstance(val, (list, tuple)):
                                print('GuiManagerAnnotations invalid input: (xytext)', val, 'should be a tuple or dict with 2 elements (coordinates)')
                        if self.textfields[j]['label'] == 'visible?' and val:
                            val = ''
                        if val != '':
                            args.update({self.textfields[j]['fromkwargs']: val})
                out['textxy'].append('')
                out['textargs'].append(args)
            i += 1
        # legend title, graph title
        keywords = ['legendtitle', 'title']
        for j in range(len(keywords)):
            tit = self.legtitvars[1][j].get().replace('\\n','\n')
            kw = stringToVariable(self.legtitvars[-1][j].get())
            if not isinstance(kw, dict):
                print('GuiManagerAnnotations invalid input:', kw, 'should be a dict (legendtitle)')
                kw = {}
            for i in range(len(self.legtitfields)):
                if 'fromkwargs' in self.legtitfields[i]:
                    if self.legtitvars[i][j].get() != '':
                        val = stringToVariable(self.legtitvars[i][j].get())
                        if 'casttype' in self.legtitfields[i]:
                            val = self.legtitfields[i]['casttype'](val)
                        kw.update({self.legtitfields[i]['fromkwargs']: val})
            if keywords[j] == 'title' and 'align' in kw:
                kw['loc'] = kw['align']
                del kw['align']
            legtit = tit if kw in ['', {}] else [tit, kw]
            out.update({keywords[j]: legtit})
        # legend properties
        kw = stringToVariable(self.legendvars[-1][0].get())
        if not isinstance(kw, dict):
            print('GuiManagerAnnotations invalid input:', kw, 'should be a dict (legend)')
            kw = {}
        for i in range(len(self.legendfields)):
            if 'fromkwargs' in self.legendfields[i]:
                if self.legendvars[i][0].get() != '':
                    val = stringToVariable(self.legendvars[i][0].get())
                    if 'casttype' in self.legendfields[i]:
                        val = self.legendfields[i]['casttype'](val)
                    kw.update({self.legendfields[i]['fromkwargs']: val})
        out.update({'legendproperties': kw})
        # graph title
        
        # perform update, finish
        #print('sent to GUI', out)
        self.funcupdate(out)
        self.refreshUI()
    def revert(self):
        self.funcupdate(self.oldattr)
        self.refreshUI()
    def refreshUI(self):
        self.fillUITextGrid(self.frameGrid)
        self.legendTitleFillValues('legendtitle', self.legtitfields, self.legtitvars)
    def close_windows(self):
        self.master.destroy()
    

    
        
        
def buildUI():
    from grapa.graph import Graph
    root = tk.Tk()
    app = GuiManagerAnnotations(root, Graph(r'.\..\examples\fancyAnnotations.txt'), None)
    app.mainloop()

if __name__ == "__main__":
    buildUI()